"""Test module for RIS 2001 api"""

from bibliograpy.api_ris2001 import Tags as RIS2001
from bibliograpy.api_ris2011 import Tags as RIS2011

def test_ris2001_retrocompatibility_support():
    """Test all RIS 2001 tags are defined in RIS 2011"""
    assert len({e for e in RIS2001 if e.name not in [t.name for t in RIS2011]}) == 0
    assert len({e for e in RIS2001 if e.name not in [t.name for t in RIS2011]}) == 0



"""Test module for bibtex api"""
import sys
import pydoc


from bibliograpy.api_bibtex import Misc, TechReport, BibtexReference
from bibliograpy.api_common import DefaultCitationFormatter
from bibliograpy.api_ris2001 import Tags as Ris2001, TypeFieldName as Ris2001Type

NASA = {Ris2001.TY: Ris2001Type.GEN, Ris2001.TI: 'NASA', Ris2001.ID: 'nasa'}

SCOPE: dict[str, BibtexReference] = {}


IAU = Misc.generic(cite_key='iau',
                   title='International Astronomical Union',
                   institution='IAU',
                   scope=SCOPE)


IAU_2006_B1 = TechReport.generic(
    cite_key='iau_2006_b1',
    author='',
    crossref='iau',
    title='Adoption of the P03 Precession Theory and Definition of the Ecliptic',
    year=2006,
    scope=SCOPE)


def test_parameterized_default_citation_formatter():
    """test parameterized default citation formatter"""

    class _DefaultCitationFormatter(DefaultCitationFormatter):
        def bibtex(self, r: BibtexReference) -> str:
            base = f'{r.title} [{r.cite_key}]'
            if r.crossref:
                return base + f' -> [{r.crossref}]'
            return base

    ref = _DefaultCitationFormatter(prefix='Références bibliographiques :',
                                    itemize='++')

    @ref(IAU_2006_B1, IAU, NASA)
    def tatafr():
        """ma doc avec plusieurs références en varargs"""


    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(tatafr) ==
"""Python Library Documentation: function tatafr in module test_api_common

t\bta\bat\bta\baf\bfr\br()
    ma doc avec plusieurs références en varargs

    Références bibliographiques :

    ++ Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1] -> [iau]
    ++ International Astronomical Union [iau]
    ++ NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(tatafr) ==
"""Python Library Documentation: function tatafr in module test_api_common

t\bta\bat\bta\baf\bfr\br()
    ma doc avec plusieurs références en varargs
    
    Références bibliographiques :
    
    ++ Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1] -> [iau]
    ++ International Astronomical Union [iau]
    ++ NASA [nasa]
""")



"""
endnote specification model.
"""
from dataclasses import dataclass
from enum import Enum, auto


@dataclass(frozen=True)
class Tag:
    """A field tag."""
    auto: auto
    _name: str | None = None
    repeating: bool = False

class Tags(Tag, Enum):
    """
    Endnote fields.
    """

    ZERO = (auto(), '0')
    """the entry type"""

    T = auto()
    """title"""

    A = (auto(), None, True)
    """author"""

    J = auto()
    """journal title"""

    V = auto()
    """volume number"""

    N = auto()
    """issue number"""

    P = auto()
    """page"""

    AT = (auto(), '@')
    """identification number (ISSN)"""

    D = auto()
    """date"""

    I = auto()
    """editor"""

    U = auto()
    """url"""

    def endnote_name(self):
        """Gets the endnote format canonical name."""
        return self._name if self._name is not None else self.name

    @staticmethod
    def parse(tag_str: str):
        """Parses a tag name into an enum value."""
        for n in Tags:
            if tag_str in ('%' + n.endnote_name(), n.endnote_name()):
                return n
        raise ValueError(f'unknown {tag_str} tag')

def default_endnote_formatter(r: dict[Tags, str | list[str]]):
    """The default formatter for endnote references."""
    title = r[Tags.T] if Tags.T in r else ""
    return f"{title}"



"""Test module for ris2011 process tool."""

from argparse import Namespace
from pathlib import Path
import pydoc
import sys
import time

import yaml

import pytest

from bibliograpy.api_common import cite

from bibliograpy.process import _process


def _ris2001_resource(file: str) -> str:
    """Chemin vers les fichiers d'entrée."""
    return str(Path(Path(__file__).parent / 'resources' / 'ris2001' / file))


def _sibbling_module(file: str) -> str:
    """Chemin vers les fichiers de modules voisins."""
    return str(Path(Path(__file__).parent / file))


def _output(file: str) -> str:
    """Chemin vers les fichiers de sortie."""
    return str(Path(Path(__file__).parent / 'resources' / 'ris2011' / 'out' / file))


def test_ris2011_yml_to_yml():
    """test process from a yml bibliography to a yml bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('sample1.yml'),
                       output_file=_output('test_ris2011_yml_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_ris2011_yml_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [
            {"TY": "JOUR",
             "A1": ["Baldwin,S.A.", "Fugaccia,I.", "Brown,D.R.", "Brown,L.V.", "Scheff,S.W."],
             "T1": "Blood-brain barrier breach following cortical contusion in the rat",
             "JO": "J.Neurosurg.",
             "Y1": "1996",
             "VL": "85",
             "SP": "476",
             "EP": "481",
             "RP": "Not In File",
             "KW": ["cortical contusion", "blood-brain barrier", "horseradish peroxidase", "head trauma",
                    "hippocampus", "rat"],
             "N2": "Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex above the "
                   "hippocampus at 3.5 m/sec with a 2 mm cortical depression. This caused severe cortical damage and "
                   "neuronal loss in hippocampus subfields CA1, CA3 and hilus. Breakdown of the blood-brain barrier "
                   "(BBB) was assessed by injecting the protein horseradish peroxidase (HRP) 5 minutes prior to or at "
                   "various times following injury (5 minutes, 1, 2, 6, 12 hours, 1, 2, 5, and 10 days). Animals were "
                   "killed 1 hour after HRP injection and brain sections were reacted with diaminobenzidine to "
                   "visualize extravascular accumulation of the protein. Maximum staining occurred in animals injected "
                   "with HRP 5 minutes prior to or 5 minutes after cortical contusion. Staining at these time points "
                   "was observed in the ipsilateral hippocampus. Some modest staining occurred in the dorsal "
                   "contralateral cortex near the superior sagittal sinus. Cortical HRP stain gradually decreased at "
                   "increasing time intervals postinjury. By 10 days, no HRP stain was observed in any area of the "
                   "brain. In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and remained so "
                   "at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the ipsilateral "
                   "hippocampus 1 and 2 days following cortical contusion, indicating a biphasic opening of the BBB "
                   "following head trauma and a possible second wave of secondary brain damage days after the "
                   "contusion injury. These data indicate regions not initially destroyed by cortical impact, but "
                   "evidencing BBB breach, may be accessible to neurotrophic factors administered intravenously both "
                   "immediately and days after brain trauma."}]

def test_ris2011_yml_to_ris():
    """test process from a yml bibliography to a RIS 2001 bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('sample1.yml'),
                       output_file=_output('test_ris2011_yml_to_ris.ris'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_ris2011_yml_to_ris.ris'), encoding='utf-8') as s:
        content = s.read()
        assert content == """TY  - JOUR
A1  - Baldwin,S.A.
A1  - Fugaccia,I.
A1  - Brown,D.R.
A1  - Brown,L.V.
A1  - Scheff,S.W.
T1  - Blood-brain barrier breach following cortical contusion in the rat
JO  - J.Neurosurg.
Y1  - 1996
VL  - 85
SP  - 476
EP  - 481
RP  - Not In File
KW  - cortical contusion
KW  - blood-brain barrier
KW  - horseradish peroxidase
KW  - head trauma
KW  - hippocampus
KW  - rat
N2  - Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex above the hippocampus at 3.5 \
m/sec with a 2 mm cortical depression. This caused severe cortical damage and neuronal loss in hippocampus subfields \
CA1, CA3 and hilus. Breakdown of the blood-brain barrier (BBB) was assessed by injecting the protein horseradish \
peroxidase (HRP) 5 minutes prior to or at various times following injury (5 minutes, 1, 2, 6, 12 hours, 1, 2, 5, and \
10 days). Animals were killed 1 hour after HRP injection and brain sections were reacted with diaminobenzidine to \
visualize extravascular accumulation of the protein. Maximum staining occurred in animals injected with HRP 5 minutes \
prior to or 5 minutes after cortical contusion. Staining at these time points was observed in the ipsilateral \
hippocampus. Some modest staining occurred in the dorsal contralateral cortex near the superior sagittal sinus. \
Cortical HRP stain gradually decreased at increasing time intervals postinjury. By 10 days, no HRP stain was observed \
in any area of the brain. In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and remained so \
at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the ipsilateral hippocampus 1 and 2 \
days following cortical contusion, indicating a biphasic opening of the BBB following head trauma and a possible \
second wave of secondary brain damage days after the contusion injury. These data indicate regions not initially \
destroyed by cortical impact, but evidencing BBB breach, may be accessible to neurotrophic factors administered \
intravenously both immediately and days after brain trauma.
ER  - 
"""

def test_ris2011_yml_to_json():
    """test process from a yml bibliography to a RIS 2001 bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('sample1.yml'),
                       output_file=_output('test_ris2011_yml_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_ris2011_yml_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"TY": "JOUR", '
                           '"A1": ["Baldwin,S.A.", "Fugaccia,I.", "Brown,D.R.", "Brown,L.V.", "Scheff,S.W."], '
                           '"T1": "Blood-brain barrier breach following cortical contusion in the rat", '
                           '"JO": "J.Neurosurg.", '
                           '"Y1": "1996", '
                           '"VL": "85", '
                           '"SP": "476", '
                           '"EP": "481", '
                           '"RP": "Not In File", '
                           '"KW": ["cortical contusion", "blood-brain barrier", "horseradish peroxidase", '
                           '"head trauma", "hippocampus", "rat"], '
                           '"N2": "Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex '
                           'above the hippocampus at 3.5 m/sec with a 2 mm cortical depression. This caused severe '
                           'cortical damage and neuronal loss in hippocampus subfields CA1, CA3 and hilus. Breakdown '
                           'of the blood-brain barrier (BBB) was assessed by injecting the protein horseradish '
                           'peroxidase (HRP) 5 minutes prior to or at various times following injury (5 minutes, 1, 2, '
                           '6, 12 hours, 1, 2, 5, and 10 days). Animals were killed 1 hour after HRP injection and '
                           'brain sections were reacted with diaminobenzidine to visualize extravascular accumulation '
                           'of the protein. Maximum staining occurred in animals injected with HRP 5 minutes prior to '
                           'or 5 minutes after cortical contusion. Staining at these time points was observed in the '
                           'ipsilateral hippocampus. Some modest staining occurred in the dorsal contralateral cortex '
                           'near the superior sagittal sinus. Cortical HRP stain gradually decreased at increasing '
                           'time intervals postinjury. By 10 days, no HRP stain was observed in any area of the brain. '
                           'In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and remained so '
                           'at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the '
                           'ipsilateral hippocampus 1 and 2 days following cortical contusion, indicating a biphasic '
                           'opening of the BBB following head trauma and a possible second wave of secondary brain '
                           'damage days after the contusion injury. These data indicate regions not initially '
                           'destroyed by cortical impact, but evidencing BBB breach, may be accessible to '
                           'neurotrophic factors administered intravenously both immediately and days after brain '
                           'trauma."}]')

def test_ris2011_json_to_yml():
    """test process from a json bibliography to a yml bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('sample1.json'),
                       output_file=_output('test_ris2011_json_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_ris2011_json_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [
            {"TY": "JOUR",
             "A1": ["Baldwin,S.A.", "Fugaccia,I.", "Brown,D.R.", "Brown,L.V.", "Scheff,S.W."],
             "T1": "Blood-brain barrier breach following cortical contusion in the rat",
             "JO": "J.Neurosurg.",
             "Y1": "1996",
             "VL": "85",
             "SP": "476",
             "EP": "481",
             "RP": "Not In File",
             "KW": ["cortical contusion", "blood-brain barrier", "horseradish peroxidase", "head trauma",
                    "hippocampus", "rat"],
             "N2": "Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex above the "
                   "hippocampus at 3.5 m/sec with a 2 mm cortical depression. This caused severe cortical damage and "
                   "neuronal loss in hippocampus subfields CA1, CA3 and hilus. Breakdown of the blood-brain barrier "
                   "(BBB) was assessed by injecting the protein horseradish peroxidase (HRP) 5 minutes prior to or at "
                   "various times following injury (5 minutes, 1, 2, 6, 12 hours, 1, 2, 5, and 10 days). Animals were "
                   "killed 1 hour after HRP injection and brain sections were reacted with diaminobenzidine to "
                   "visualize extravascular accumulation of the protein. Maximum staining occurred in animals injected "
                   "with HRP 5 minutes prior to or 5 minutes after cortical contusion. Staining at these time points "
                   "was observed in the ipsilateral hippocampus. Some modest staining occurred in the dorsal "
                   "contralateral cortex near the superior sagittal sinus. Cortical HRP stain gradually decreased at "
                   "increasing time intervals postinjury. By 10 days, no HRP stain was observed in any area of the "
                   "brain. In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and remained so "
                   "at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the ipsilateral "
                   "hippocampus 1 and 2 days following cortical contusion, indicating a biphasic opening of the BBB "
                   "following head trauma and a possible second wave of secondary brain damage days after the "
                   "contusion injury. These data indicate regions not initially destroyed by cortical impact, but "
                   "evidencing BBB breach, may be accessible to neurotrophic factors administered intravenously both "
                   "immediately and days after brain trauma."}]

def test_ris2011_json_to_ris():
    """test process from a json bibliography to a RIS 2001 bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('sample1.json'),
                       output_file=_output('test_ris2011_json_to_ris.ris'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_ris2011_json_to_ris.ris'), encoding='utf-8') as s:
        content = s.read()
        assert content == """TY  - JOUR
A1  - Baldwin,S.A.
A1  - Fugaccia,I.
A1  - Brown,D.R.
A1  - Brown,L.V.
A1  - Scheff,S.W.
T1  - Blood-brain barrier breach following cortical contusion in the rat
JO  - J.Neurosurg.
Y1  - 1996
VL  - 85
SP  - 476
EP  - 481
RP  - Not In File
KW  - cortical contusion
KW  - blood-brain barrier
KW  - horseradish peroxidase
KW  - head trauma
KW  - hippocampus
KW  - rat
N2  - Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex above the hippocampus at 3.5 \
m/sec with a 2 mm cortical depression. This caused severe cortical damage and neuronal loss in hippocampus subfields \
CA1, CA3 and hilus. Breakdown of the blood-brain barrier (BBB) was assessed by injecting the protein horseradish \
peroxidase (HRP) 5 minutes prior to or at various times following injury (5 minutes, 1, 2, 6, 12 hours, 1, 2, 5, and \
10 days). Animals were killed 1 hour after HRP injection and brain sections were reacted with diaminobenzidine to \
visualize extravascular accumulation of the protein. Maximum staining occurred in animals injected with HRP 5 minutes \
prior to or 5 minutes after cortical contusion. Staining at these time points was observed in the ipsilateral \
hippocampus. Some modest staining occurred in the dorsal contralateral cortex near the superior sagittal sinus. \
Cortical HRP stain gradually decreased at increasing time intervals postinjury. By 10 days, no HRP stain was observed \
in any area of the brain. In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and remained so \
at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the ipsilateral hippocampus 1 and 2 \
days following cortical contusion, indicating a biphasic opening of the BBB following head trauma and a possible \
second wave of secondary brain damage days after the contusion injury. These data indicate regions not initially \
destroyed by cortical impact, but evidencing BBB breach, may be accessible to neurotrophic factors administered \
intravenously both immediately and days after brain trauma.
ER  - 
"""

def test_ris2011_json_to_json():
    """test process from a json bibliography to a json bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('sample1.json'),
                       output_file=_output('test_ris2011_json_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_ris2011_json_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"TY": "JOUR", '
                           '"A1": ["Baldwin,S.A.", "Fugaccia,I.", "Brown,D.R.", "Brown,L.V.", "Scheff,S.W."], '
                           '"T1": "Blood-brain barrier breach following cortical contusion in the rat", '
                           '"JO": "J.Neurosurg.", '
                           '"Y1": "1996", '
                           '"VL": "85", '
                           '"SP": "476", '
                           '"EP": "481", '
                           '"RP": "Not In File", '
                           '"KW": ["cortical contusion", "blood-brain barrier", "horseradish peroxidase", '
                           '"head trauma", "hippocampus", "rat"], '
                           '"N2": "Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex '
                           'above the hippocampus at 3.5 m/sec with a 2 mm cortical depression. This caused severe '
                           'cortical damage and neuronal loss in hippocampus subfields CA1, CA3 and hilus. Breakdown '
                           'of the blood-brain barrier (BBB) was assessed by injecting the protein horseradish '
                           'peroxidase (HRP) 5 minutes prior to or at various times following injury (5 minutes, 1, 2, '
                           '6, 12 hours, 1, 2, 5, and 10 days). Animals were killed 1 hour after HRP injection and '
                           'brain sections were reacted with diaminobenzidine to visualize extravascular accumulation '
                           'of the protein. Maximum staining occurred in animals injected with HRP 5 minutes prior to '
                           'or 5 minutes after cortical contusion. Staining at these time points was observed in the '
                           'ipsilateral hippocampus. Some modest staining occurred in the dorsal contralateral cortex '
                           'near the superior sagittal sinus. Cortical HRP stain gradually decreased at increasing '
                           'time intervals postinjury. By 10 days, no HRP stain was observed in any area of the brain. '
                           'In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and remained so '
                           'at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the '
                           'ipsilateral hippocampus 1 and 2 days following cortical contusion, indicating a biphasic '
                           'opening of the BBB following head trauma and a possible second wave of secondary brain '
                           'damage days after the contusion injury. These data indicate regions not initially '
                           'destroyed by cortical impact, but evidencing BBB breach, may be accessible to '
                           'neurotrophic factors administered intravenously both immediately and days after brain '
                           'trauma."}]')

def test_ris2011_ris_to_yml():
    """test process from a bib bibliography to a yml bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('sample1.ris'),
                       output_file=_output('test_ris2011_ris_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_ris2011_ris_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [
            {"TY": "JOUR",
             "A1": ["Baldwin,S.A.", "Fugaccia,I.", "Brown,D.R.", "Brown,L.V.", "Scheff,S.W."],
             "T1": "Blood-brain barrier breach following cortical contusion in the rat",
             "JO": "J.Neurosurg.",
             "Y1": "1996",
             "VL": "85",
             "SP": "476",
             "EP": "481",
             "RP": "Not In File",
             "KW": ["cortical contusion", "blood-brain barrier", "horseradish peroxidase", "head trauma",
                    "hippocampus", "rat"],
             "N2": "Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex above the "
                   "hippocampus at 3.5 m/sec with a 2 mm cortical depression. This caused severe cortical damage and "
                   "neuronal loss in hippocampus subfields CA1, CA3 and hilus. Breakdown of the blood-brain barrier "
                   "(BBB) was assessed by injecting the protein horseradish peroxidase (HRP) 5 minutes prior to or at "
                   "various times following injury (5 minutes, 1, 2, 6, 12 hours, 1, 2, 5, and 10 days). Animals were "
                   "killed 1 hour after HRP injection and brain sections were reacted with diaminobenzidine to "
                   "visualize extravascular accumulation of the protein. Maximum staining occurred in animals injected "
                   "with HRP 5 minutes prior to or 5 minutes after cortical contusion. Staining at these time points "
                   "was observed in the ipsilateral hippocampus. Some modest staining occurred in the dorsal "
                   "contralateral cortex near the superior sagittal sinus. Cortical HRP stain gradually decreased at "
                   "increasing time intervals postinjury. By 10 days, no HRP stain was observed in any area of the "
                   "brain. In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and remained so "
                   "at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the ipsilateral "
                   "hippocampus 1 and 2 days following cortical contusion, indicating a biphasic opening of the BBB "
                   "following head trauma and a possible second wave of secondary brain damage days after the "
                   "contusion injury. These data indicate regions not initially destroyed by cortical impact, but "
                   "evidencing BBB breach, may be accessible to neurotrophic factors administered intravenously both "
                   "immediately and days after brain trauma."}]

def test_ris2011_ris_to_ris():
    """test process from a RIS 2001 bibliography to a RIS 2001 bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('sample1.ris'),
                       output_file=_output('test_ris2011_ris_to_ris.ris'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_ris2011_ris_to_ris.ris'), encoding='utf-8') as s:
        content = s.read()
        assert content == """TY  - JOUR
A1  - Baldwin,S.A.
A1  - Fugaccia,I.
A1  - Brown,D.R.
A1  - Brown,L.V.
A1  - Scheff,S.W.
T1  - Blood-brain barrier breach following cortical contusion in the rat
JO  - J.Neurosurg.
Y1  - 1996
VL  - 85
SP  - 476
EP  - 481
RP  - Not In File
KW  - cortical contusion
KW  - blood-brain barrier
KW  - horseradish peroxidase
KW  - head trauma
KW  - hippocampus
KW  - rat
N2  - Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex above the hippocampus at 3.5 \
m/sec with a 2 mm cortical depression. This caused severe cortical damage and neuronal loss in hippocampus subfields \
CA1, CA3 and hilus. Breakdown of the blood-brain barrier (BBB) was assessed by injecting the protein horseradish \
peroxidase (HRP) 5 minutes prior to or at various times following injury (5 minutes, 1, 2, 6, 12 hours, 1, 2, 5, and \
10 days). Animals were killed 1 hour after HRP injection and brain sections were reacted with diaminobenzidine to \
visualize extravascular accumulation of the protein. Maximum staining occurred in animals injected with HRP 5 minutes \
prior to or 5 minutes after cortical contusion. Staining at these time points was observed in the ipsilateral \
hippocampus. Some modest staining occurred in the dorsal contralateral cortex near the superior sagittal sinus. \
Cortical HRP stain gradually decreased at increasing time intervals postinjury. By 10 days, no HRP stain was observed \
in any area of the brain. In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and remained so \
at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the ipsilateral hippocampus 1 and 2 \
days following cortical contusion, indicating a biphasic opening of the BBB following head trauma and a possible \
second wave of secondary brain damage days after the contusion injury. These data indicate regions not initially \
destroyed by cortical impact, but evidencing BBB breach, may be accessible to neurotrophic factors administered \
intravenously both immediately and days after brain trauma.
ER  - 
"""

def test_ris2011_ris_to_json():
    """test process from a RIS 2001 bibliography to a json bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('sample1.ris'),
                       output_file=_output('test_ris2011_ris_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_ris2011_ris_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"TY": "JOUR", '
                           '"A1": ["Baldwin,S.A.", "Fugaccia,I.", "Brown,D.R.", "Brown,L.V.", "Scheff,S.W."], '
                           '"T1": "Blood-brain barrier breach following cortical contusion in the rat", '
                           '"JO": "J.Neurosurg.", '
                           '"Y1": "1996", '
                           '"VL": "85", '
                           '"SP": "476", '
                           '"EP": "481", '
                           '"RP": "Not In File", '
                           '"KW": ["cortical contusion", "blood-brain barrier", "horseradish peroxidase", '
                           '"head trauma", "hippocampus", "rat"], '
                           '"N2": "Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex '
                           'above the hippocampus at 3.5 m/sec with a 2 mm cortical depression. This caused severe '
                           'cortical damage and neuronal loss in hippocampus subfields CA1, CA3 and hilus. Breakdown '
                           'of the blood-brain barrier (BBB) was assessed by injecting the protein horseradish '
                           'peroxidase (HRP) 5 minutes prior to or at various times following injury (5 minutes, 1, 2, '
                           '6, 12 hours, 1, 2, 5, and 10 days). Animals were killed 1 hour after HRP injection and '
                           'brain sections were reacted with diaminobenzidine to visualize extravascular accumulation '
                           'of the protein. Maximum staining occurred in animals injected with HRP 5 minutes prior to '
                           'or 5 minutes after cortical contusion. Staining at these time points was observed in the '
                           'ipsilateral hippocampus. Some modest staining occurred in the dorsal contralateral cortex '
                           'near the superior sagittal sinus. Cortical HRP stain gradually decreased at increasing '
                           'time intervals postinjury. By 10 days, no HRP stain was observed in any area of the brain. '
                           'In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and remained so '
                           'at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the '
                           'ipsilateral hippocampus 1 and 2 days following cortical contusion, indicating a biphasic '
                           'opening of the BBB following head trauma and a possible second wave of secondary brain '
                           'damage days after the contusion injury. These data indicate regions not initially '
                           'destroyed by cortical impact, but evidencing BBB breach, may be accessible to '
                           'neurotrophic factors administered intravenously both immediately and days after brain '
                           'trauma."}]')

def test_ris2011_yml_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('mini.yml'),
                       output_file=_sibbling_module('test_ris2011_yml_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_ris2011_yml_to_py import IAU, NASA

    @cite(IAU, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_ris2011

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union [iau]
    * NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_ris2011

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union [iau]
    * NASA [nasa]
""")

def test_ris2011_json_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('mini.json'),
                       output_file=_sibbling_module('test_ris2011_json_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_ris2011_json_to_py import IAU, NASA

    @cite(IAU, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_ris2011

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union [iau]
    * NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_ris2011

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union [iau]
    * NASA [nasa]
""")

def test_ris2011_ris_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='ris2011',
                       file=_ris2001_resource('mini.ris'),
                       output_file=_sibbling_module('test_ris2011_ris_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_ris2011_ris_to_py import IAU, NASA

    @cite(IAU, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_ris2011

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union [iau]
    * NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_ris2011

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union [iau]
    * NASA [nasa]
""")

def test_input_file_not_found():
    """test process input file not found"""

    with pytest.raises(FileNotFoundError) as e:
        with open(_ris2001_resource('not_existing_file.yml'), encoding='utf-8') as s:
            yaml.safe_load(s)

    assert e.value.args[0] == 2
    assert e.value.args[1] == "No such file or directory"



"""Test module for refer IO"""
from pathlib import Path
from typing import TextIO

from bibliograpy.api_core import Formats
from bibliograpy.api_refer import Tags
from bibliograpy.io_refer import ReferInputFormat


def read_entries(i: TextIO):
    """ris entry parsing shortcut"""
    return ReferInputFormat(Formats.REFER).from_standard(i)

def test_sample1():
    """test refer sample"""

    with open(Path(__file__).parent / 'resources' / 'refer' / 'sample1.refer', encoding='utf-8') as s:
        result = read_entries(s)
        assert len(result) == 1
        assert result[0] == {
            Tags.A: ['Brian W. Kernighan', 'Lorinda L. Cherry'],
            Tags.T: 'A System for Typesetting Mathematics',
            Tags.J: 'J. Comm. ACM',
            Tags.V: '18',
            Tags.N: '3',
            Tags.D: 'March 1978',
            Tags.P: '151-157',
            Tags.K: 'eqn'
        }

def test_sample2():
    """test refer sample"""

    with open(Path(__file__).parent / 'resources' / 'refer' / 'sample2.refer', encoding='utf-8') as s:
        result = read_entries(s)
        assert len(result) == 1
        assert result[0] == {
            Tags.A: ['Brian W. Kernighan', 'Lorinda L. Cherry'],
            Tags.T: 'A System for Typesetting Mathematics',
            Tags.J: 'J. Comm. ACM',
            Tags.V: '18',
            Tags.N: '3',
            Tags.D: 'March 1978',
            Tags.P: '151-157',
            Tags.K: 'eqn'
        }

def test_sample3():
    """test refer sample"""

    with open(Path(__file__).parent / 'resources' / 'refer' / 'sample3.refer', encoding='utf-8') as s:
        result = read_entries(s)
        assert len(result) == 2
        assert result[0] == {
            Tags.A: ['Brian W. Kernighan'],
            Tags.T: 'A System for Typesetting Mathematics',
            Tags.J: 'J. Comm. ACM',
            Tags.V: '18',
            Tags.N: '3',
            Tags.D: 'March 1978',
            Tags.P: '151-157',
            Tags.K: 'eqn'
        }
        assert result[1] == {
            Tags.A: ['Lorinda L. Cherry'],
            Tags.T: 'A System for Typesetting Mathematics',
            Tags.J: 'J. Comm. ACM',
            Tags.V: '18',
            Tags.N: '3',
            Tags.D: 'March 1978',
            Tags.P: '151-157',
            Tags.K: 'eqn'
        }



"""
bibliograpy command entrypoint
"""
import logging

from argparse import ArgumentParser, Namespace

from bibliograpy.api_core import Formats
from bibliograpy.process import _process

LOG = logging.getLogger(__name__)


DEFAULT_FILE = "bibliograpy.yaml"
DEFAULT_ENCODING = 'utf-8'
DEFAULT_OUTPUT_DIR = '.'
DEFAULT_OUTPUT_FILE = 'bibliography.py'
DEFAULT_FORMAT = 'bib'
DEFAULT_INIT_SCOPE = '{}'


def _create_parser() -> ArgumentParser:

    # parse argument line
    parser = ArgumentParser(description='Bibliography management.')

    subparsers = parser.add_subparsers(dest='CMD', help='available commands')

    bibtex = subparsers.add_parser(name=Formats.BIBTEX.command,
                                   help='generates bibliograpy Python source bibliography '
                                        f'from {Formats.BIBTEX.title} format')
    bibtex.add_argument('file',
                         nargs='?',
                         help="path to the input bibliography file",
                         default=DEFAULT_FILE)
    bibtex.add_argument('--encoding', '-e',
                         nargs='?',
                         help='the bibliograpy configuration file encoding (default to utf-8)',
                         default=DEFAULT_ENCODING)
    bibtex.add_argument('--output-dir', '-O',
                         nargs='?',
                         help='the source bibliograpy file output directory',
                         default=DEFAULT_OUTPUT_DIR)
    bibtex.add_argument('--output-file', '-o',
                         nargs='?',
                         help='the source bibliograpy output file name',
                         default=DEFAULT_OUTPUT_FILE)
    group = bibtex.add_mutually_exclusive_group()
    group.add_argument('--scope', '-s',
                       nargs='?',
                       help="""the scope name, must be consistent with --init-scope \
    (for bibtex format cross-reference resolution)""")
    group.add_argument('--shared-scope', '-S',
                       action='store_true',
                       help='use the default shared scope named SHARED_SCOPE')
    bibtex.add_argument('--init-scope', '-i',
                         nargs='?',
                         help='the scope initialization value line (for bibtex format cross-reference resolution)')


    for fmt in [f for f in Formats if f.command is not None and f is not Formats.BIBTEX]:

        f = subparsers.add_parser(name=fmt.command,
                                  help=f'generates bibliograpy Python source bibliography from {fmt.title} format')
        f.add_argument('file',
                       nargs='?',
                       help="path to the input bibliography file",
                       default=DEFAULT_FILE)
        f.add_argument('--encoding', '-e',
                       nargs='?',
                       help='the bibliograpy configuration file encoding (default to utf-8)',
                       default=DEFAULT_ENCODING)
        f.add_argument('--output-dir', '-O',
                       nargs='?',
                       help='the source bibliograpy file output directory',
                       default=DEFAULT_OUTPUT_DIR)
        f.add_argument('--output-file', '-o',
                       nargs='?',
                       help='the source bibliograpy output file name',
                       default=DEFAULT_OUTPUT_FILE)

    return parser


def entrypoint():
    """The pyenvs command entrypoint."""

    commands = { f.command: _process for f in Formats if f.command is not None }

    ns: Namespace = _create_parser().parse_args()

    commands.get(ns.CMD)(ns)



"""
bibliograpy process module
"""
import logging
from argparse import Namespace
from pathlib import Path

from bibliograpy.api_core import Formats
from bibliograpy.io_bibtex import BibtexInputFormat, BibtexOutputFormat
from bibliograpy.io_refer import ReferInputFormat, ReferOutputFormat
from bibliograpy.io_ris2001 import Ris2001InputFormat, Ris2001OutputFormat
from bibliograpy.io_ris2011 import Ris2011InputFormat, Ris2011OutputFormat
from bibliograpy.io_endnote import EndnoteInputFormat, EndnoteOutputFormat

LOG = logging.getLogger(__name__)


def _process(ns: Namespace):
    """config
    """
    LOG.info("dependencies")

    source = Formats.as_io_extension(ns.file.split('.')[-1])
    output_dir = Path(Path.cwd(), ns.output_dir)
    output_file = ns.output_file
    target = Formats.as_io_extension(output_file.split('.')[-1])
    if 'shared_scope' in ns and ns.shared_scope:
        scope_symbol = 'SHARED_SCOPE'
    else:
        scope_symbol = ns.scope if 'scope' in ns else None
    init_scope = ns.init_scope if 'init_scope' in ns else None
    fmt = Formats.as_command(ns.CMD)



    LOG.info('open configuration file %s', ns.file)

    if fmt is Formats.BIBTEX:
        iformat = BibtexInputFormat(source=source)
        with open(ns.file, encoding=ns.encoding) as i:
            content = iformat.read(i)
            oformat = BibtexOutputFormat(content=content,
                                         target=target,
                                         scope_symbol=scope_symbol,
                                         init_scope=init_scope)
            with open(Path(output_dir, output_file), 'w', encoding=ns.encoding) as o:
                oformat.write(o)

    elif fmt is Formats.RIS2001:
        iformat = Ris2001InputFormat(source=source)
        with open(ns.file, encoding=ns.encoding) as i:
            content = iformat.read(i)
            oformat = Ris2001OutputFormat(target=target, content=content)
            with open(Path(output_dir, output_file), 'w', encoding=ns.encoding) as o:
                oformat.write(o)

    elif fmt is Formats.RIS2011:
        iformat = Ris2011InputFormat(source=source)
        with open(ns.file, encoding=ns.encoding) as i:
            content = iformat.read(i)
            oformat = Ris2011OutputFormat(target=target, content=content)
            with open(Path(output_dir, output_file), 'w', encoding=ns.encoding) as o:
                oformat.write(o)

    elif fmt is Formats.REFER:
        iformat = ReferInputFormat(source=source)
        with open(ns.file, encoding=ns.encoding) as i:
            content = iformat.read(i)
            oformat = ReferOutputFormat(target=target, content=content)
            with open(Path(output_dir, output_file), 'w', encoding=ns.encoding) as o:
                oformat.write(o)

    elif fmt is Formats.ENDNOTE:
        iformat = EndnoteInputFormat(source=source)
        with open(ns.file, encoding=ns.encoding) as i:
            content = iformat.read(i)
            oformat = EndnoteOutputFormat(target=target, content=content)
            with open(Path(output_dir, output_file), 'w', encoding=ns.encoding) as o:
                oformat.write(o)
    else:
        raise ValueError(f'unsupported format {format}')



"""
RIS 2001 specification model.
"""
from dataclasses import dataclass
from enum import Enum, auto

from bibliograpy.bibliography import RIS_2001
from bibliograpy.api_bibtex import _cite


@dataclass(frozen=True)
class Tag:
    """A field tag."""
    auto: auto
    repeating: bool = False

@_cite(RIS_2001)
class Tags(Tag, Enum):
    """
    RIS fields.
    """

    TY = auto()
    """Type of reference. 
    This must contain one of the following field names as defined in the section, Reference Type field names."""

    ER = auto()
    """End of reference.
    Must be the last tag in a reference."""

    ID = auto()
    """Reference ID.
    The Reference ID can consist of any alphanumeric characters—up to 20 characters in length."""

    T1 = auto()
    """Title Primary.
    Note that the BT tag maps to this field only for Whole Book and Unpublished Work references.
    This field can contain alphanumeric characters; there is no practical length limit to this field."""
    TI = auto()  # synonym of T1
    CT = auto()  # synonym of T1

    BT = auto()

    T2 = auto()
    """Title Secondary.
    Note that the BT tag maps to this field for all reference types except for Whole Book and Unpublished Work 
    references.
    There is no practical limit to the length of this field."""

    T3 = auto()
    """Title Series.
    This field can contain alphanumeric characters; there is no practical length limit to this field."""

    A1 = (auto(), True)
    """Author Primary.
    Each author must be on a separate line, preceded by this tag. Each reference can contain unlimited author fields, 
    and can contain up to 255 characters for each field. The author name must be in the following syntax:

    Lastname, Firstname, Suffix

    For Firstname, you can use full names, initials, or both. The format for the author’s first name is as follows:

    Phillips,A.J

    Phillips,Albert John

    Phillips,Albert

    Lastname = Any string of letters, spaces, and hyphens

    Firstname = Any string of letters, spaces, and hyphens

    Initial = Any single letter followed by a period

    Full Name = Any string of letters, spaces, and hyphens

    Suffix = Jr/Sr/II/III/MD etc. (Phillips,A.J.,Sr.); use of the suffix is optional"""
    AU = (auto(), True)  # synonym of A1

    A2 = (auto(), True)
    """Author Secondary.
    Each author must be on a separate line, preceded by this tag. There is no practical limit to the number of authors
    in this field. The author name must be in the correct syntax (refer to A1 and AU fields).
    This author name can be up to 255 characters long."""
    ED = (auto(), True)  # synonym of A2

    A3 = (auto(), True)
    """Author Series. 
	Each author must be on a separate line, preceded by this tag. There is no practical limit to the number of authors 
	in this field. The author name must be in the correct syntax (refer to A1 and AU fields).
	Each author name can be up to 255 characters long."""

    Y1 = auto()
    """Date Primary.
    This date must be in the following format:

    YYYY/MM/DD/other info

    The year, month and day fields are all numeric. The other info field can be any string of letters, spaces and
    hyphens. Note that each specific date information is optional, however the slashes ("/") are not. For example, if
    you just had the <year> and <other info>, then the output would look like: "1998///Spring."
    """
    PY = auto()  # synonym of Y1

    Y2 = auto()
    """Date Secondary. (Refer to Y1 and PY fields)."""

    N1 = auto()
    """Notes.
    These are free text fields and can contain alphanumeric characters; there is no practical length limit to this
    field."""
    AB = auto()  # synonym of Y1

    N2 = auto()
    """Abstract.
    This is a free text field and can contain alphanumeric characters; there is no practical length limit to this field.
    """

    KW = (auto(), True)
    """Keywords.
    Each keyword or phrase must be on its own line, preceded by this tag. A keyword can consist of multiple words
    (phrases) and can be up to 255 characters long. There is no limit to the amount of keywords in a single reference.
    """

    RP = auto()
    """Reprint status.
    This optional field can contain one of three status notes. Each must be in uppercase,
    and the date after "ON REQUEST" must be in the US date format, in parentheses: (MM/DD/YY). 
    If this field is blank in your downloaded text file, the import function assumes the reprint status is 
    “NOT IN FILE.”
    
    The three options are:

    IN FILE - This is for references that you have a physical copy of in your files.
    NOT IN FILE - This is for references that you do not have physical copies of in your files.
    ON REQUEST (MM/DD/YY) - This means that you have sent for a reprint of the reference; the date is the date on which
    the reprint was requested (in MM/DD/YY format)."""

    JF = auto()
    """Periodical name: full format.
    This is an alphanumeric field of up to 255 characters."""
    JO = auto()  # synonym of JF

    JA = auto()
    """Periodical name: standard abbreviation.
    This is the periodical in which the article was (or is to be, in the case of in-press references) published.
    This is an alphanumeric field of up to 255 characters.

    If possible, periodical names should be abbreviated in the Index Medicus style, with periods after the
    abbreviations. If this is not possible (your large bibliography file in WordPerfect has no periods after
    abbreviations), you can use the "RIS Format (Adds periods)" Import Filter definition. This definition uses the
    Periodical Word Dictionary."""

    J1 = auto()
    """Periodical name: user abbreviation 1.
	This is an alphanumeric field of up to 255 characters."""

    J2 = auto()
    """Periodical name: user abbreviation 2.
    This is an alphanumeric field of up to 255 characters."""

    VL = auto()
    """Volume number.
    There is no practical limit to the length of this field."""

    IS = auto()
    """Issue.
    There is no practical limit to the length of this field."""
    CP = auto()  # synonym of IS

    SP = auto()
    """Start page number; an alphanumeric string.
    There is no practical limit to the length of this field."""

    EP = auto()
    """Ending page number, as above."""

    CY = auto()
    """City of publication; this is an alphanumeric field.
    There is no practical limit to the length of this field."""

    PB = auto()
    """Publisher; this is an alphanumeric field.
    There is no practical limit to the length of this field."""

    SN = auto()
    """ISSN/ISBN. This is an alphanumeric field.
    There is no practical limit to the length of this field."""

    AD = auto()
    """Address.
    This is a free text field and contain alphanumeric characters; there is no practical length limit to this field."""

    AV = auto()
    """Availability.
    This is an alphanumeric field and there is no practical limit to the length of this field."""

    M1 = auto()
    """Miscellaneous 1.
    This is an alphanumeric field and there is no practical limit to the length of this field."""

    M2 = auto()
    """Miscellaneous 2.
    This is an alphanumeric field and there is no practical limit to the length of this field."""

    M3 = auto()
    """Miscellaneous 3.
    This is an alphanumeric field and there is no practical limit to the length of this field."""

    U1 = auto()
    """User definable 1.
    This is an alphanumeric field and there is no practical limit to the length of this field."""

    U2 = auto()
    """User definable 2.
    This is an alphanumeric field and there is no practical limit to the length of this field."""

    U3 = auto()
    """User definable 3.
    This is an alphanumeric field and there is no practical limit to the length of this field."""

    U4 = auto()
    """User definable 4.
    This is an alphanumeric field and there is no practical limit to the length of this field."""

    U5 = auto()
    """User definable 5.
    This is an alphanumeric field and there is no practical limit to the length of this field."""

    UR = auto()
    """Web/URL.
	There is no practical limit to the length of this field.
	URL addresses can be entered individually, one per tag or multiple addresses can be entered on one line
	using a semi-colon as a separator."""

    L1 = auto()
    """Link to PDF.
    There is no practical limit to the length of this field.
    URL addresses can be entered individually, one per tag or multiple addresses can be entered on one line
    using a semi-colon as a separator."""

    L2 = auto()
    """Link to Full-text.
    There is no practical limit to the length of this field.
    URL addresses can be entered individually, one per tag or multiple addresses can be entered on one line
    using a semi-colon as a separator."""

    L3 = auto()
    """Related Records.
    There is no practical limit to the length of this field."""

    L4 = auto()
    """Image(s).
    There is no practical limit to the length of this field."""

    @staticmethod
    def parse(tag_str: str):
        """Parses a tag name into an enum value."""
        for n in Tags:
            if tag_str == n.name:
                return n
        raise ValueError(f'unknown {tag_str} tag')


@_cite(RIS_2001)
class TypeFieldName(Enum):
    """Reference Type Field Names

    The following describes the valid reference type field names that can be used with for the reference type field when
    importing references into Reference Manager."""

    ABST = auto()
    """Abstract"""

    ADVS = auto()
    """Audiovisual material"""

    ART = auto()
    """Art Work"""

    BILL = auto()
    """Bill/Resolution"""

    BOOK = auto()
    """Book, Whole"""

    CASE = auto()
    """Case"""

    CHAP = auto()
    """Book chapter"""

    COMP = auto()
    """Computer program"""

    CONF = auto()
    """Conference proceeding"""

    CTLG = auto()
    """Catalog"""

    DATA = auto()
    """Data file"""

    ELEC = auto()
    """Electronic Citation"""

    GEN = auto()
    """Generic"""

    HEAR = auto()
    """Hearing"""

    ICOMM = auto()
    """Internet Communication"""

    INPR = auto()
    """In Press"""

    JFULL = auto()
    """Journal (full)"""

    JOUR = auto()
    """Journal"""

    MAP = auto()
    """Map"""

    MGZN = auto()
    """Magazine article"""

    MPCT = auto()
    """Motion picture"""

    MUSIC = auto()
    """Music score"""

    NEWS = auto()
    """Newspaper"""

    PAMP = auto()
    """Pamphlet"""

    PAT = auto()
    """Patent"""

    PCOMM = auto()
    """Personal communication"""

    RPRT = auto()
    """Report"""

    SER = auto()
    """Serial (Book, Monograph)"""

    SLIDE = auto()
    """Slide"""

    SOUND = auto()
    """Sound recording"""

    STAT = auto()
    """Statute"""

    THES = auto()
    """Thesis/Dissertation"""

    UNBILL = auto()
    """Unenacted bill/resolution"""

    UNPB = auto()
    """Unpublished work"""

    VIDEO = auto()
    """Video recording"""

    @staticmethod
    def parse(entry_type: str):
        """Parses an entry type name into an enum value."""
        for n in TypeFieldName:
            if entry_type == n.name:
                return n
        raise ValueError(f'unknown {entry_type} type')


def default_ris2001_formatter(r: dict[Tags, str | list[str] | TypeFieldName]):
    """The default formatter for RIS 2001 references."""
    title = r[Tags.TI] if Tags.TI in r else (r[Tags.T1] if Tags.T1 in r else (r[Tags.CT] if Tags.CT in r else ""))
    return f"{title} [{r[Tags.ID]}]" if Tags.ID in r else title



"""
refer specification model.
"""
from dataclasses import dataclass
from enum import Enum, auto

from bibliograpy.bibliography import REFER_MAN
from bibliograpy.api_bibtex import _cite


@dataclass(frozen=True)
class Tag:
    """A field tag."""
    auto: auto
    repeating: bool = False

@_cite(REFER_MAN)
class Tags(Tag, Enum):
    """
    REFER fields.
    """

    A = (auto(), True)
    """The name of an author.  If the name contains a title such as Jr.
    at the end, it should be separated  from  the  last  name  by  a
    comma.   There can be multiple occurrences of the %A field.  The
    order is significant.  It is a good idea always to supply an  %A
    field or a %Q field."""

    B = auto()
    """For an article that is part of a book, the title of the book."""

    C = auto()
    """The place (city) of publication."""

    D = auto()
    """The  date of publication.  The year should be specified in full.
    If the month is specified, the name rather than  the  number  of
    the  month  should be used, but only the first three letters are
    required.  It is a good idea always to supply a %D field; if the
    date  is  unknown,  a  value  such as in press or unknown can be
    used."""

    E = (auto(), True)
    """For an article that is part of a book, the name of an editor  of
    the  book.  Where the work has editors and no authors, the names
    of the editors should be  given  as  %A  fields  and  , (ed)  or
    , (eds) should be appended to the last author."""

    G = auto()
    """US Government ordering number."""

    I = auto()
    """The publisher (issuer)."""

    J = auto()
    """For an article in a journal, the name of the journal."""

    K = auto()
    """Keywords to be used for searching."""

    L = auto()
    """Label."""

    N = auto()
    """Journal issue number."""

    O = auto()
    """Other  information.   This  is usually printed at the end of the
    reference."""

    P = auto()
    """Page number.  A range of pages can be specified as m-n."""

    Q = (auto(), False)
    """The name of the author, if the author is  not  a  person.   This
    will  only be used if there are no %A fields.  There can only be
    one %Q field."""

    R = auto()
    """Technical report number."""

    S = auto()
    """Series name."""

    T = auto()
    """Title.  For an article in a book or journal, this should be  the
    title of the article."""

    V = auto()
    """Volume number of the journal or book."""

    X = auto()
    """Annotation."""

    @staticmethod
    def parse(tag_str: str):
        """Parses a tag name into an enum value."""
        for n in Tags:
            if tag_str in (n.name, "%" + n.name):
                return n
        raise ValueError(f'unknown {tag_str} tag')

def default_refer_formatter(r: dict[Tags, str | list[str]]):
    """The default formatter for refer references."""
    title = r[Tags.T] if Tags.T in r else ""
    return f"{title} [{r[Tags.L]}]" if Tags.L in r else title



"""Test module for endnote IO"""
from pathlib import Path
from typing import TextIO

from bibliograpy.api_core import Formats
from bibliograpy.api_endnote import Tags
from bibliograpy.io_endnote import EndnoteInputFormat


def read_entries(i: TextIO):
    """ris entry parsing shortcut"""
    return EndnoteInputFormat(Formats.ENDNOTE).from_standard(i)

def test_sample1():
    """test endnote sample1"""

    with open(Path(__file__).parent / 'resources' / 'endnote' / 'sample1.enw', encoding='utf-8') as s:
        result = read_entries(s)
        assert len(result) == 1
        assert result[0] == {
            Tags.D: '2015',
            Tags.I: 'Annual Reviews',
            Tags.N: '1',
            Tags.P: '163-193',
            Tags.AT: '0066-4189',
            Tags.J: 'Annual Review of Fluid Mechanics',
            Tags.V: '47',
            Tags.ZERO: 'Journal Article',
            Tags.T: 'Flows driven by libration, precession, and tides',
            Tags.A: ['Le Bars, Michael', 'Cébron, David', 'Le Gal, Patrice']
        }

def test_sample2():
    """test endnote sample2"""

    with open(Path(__file__).parent / 'resources' / 'endnote' / 'sample2.enw', encoding='utf-8') as s:
        result = read_entries(s)
        assert len(result) == 1
        assert result[0] == {
            Tags.D: '2003',
            Tags.I: 'IOP Publishing',
            Tags.N: '1',
            Tags.P: '494',
            Tags.AT: '1538-3881',
            Tags.J: 'The Astronomical Journal',
            Tags.V: '126',
            Tags.ZERO: 'Journal Article',
            Tags.T: 'A new precession formula',
            Tags.A: ['Fukushima, Toshio']
        }

def test_sample3():
    """test endnote sample3"""

    with open(Path(__file__).parent / 'resources' / 'endnote' / 'sample3.enw', encoding='utf-8') as s:
        result = read_entries(s)
        assert len(result) == 2
        assert result[0] == {
            Tags.D: '2016',
            Tags.I: 'Elsevier',
            Tags.P: '317-323',
            Tags.AT: '0096-3003',
            Tags.J: 'Applied Mathematics and Computation',
            Tags.V: '275',
            Tags.ZERO: 'Journal Article',
            Tags.T: 'A new approach on curves of constant precession',
            Tags.A: ['Uzunoğlu, Beyhan', 'Gök, İsmail', 'Yaylı, Yusuf']
        }
        assert result[1] == {
            Tags.D: '2016',
            Tags.I: 'Elsevier',
            Tags.P: '317-323',
            Tags.AT: '0096-3003',
            Tags.J: 'Applied Mathematics and Computation',
            Tags.V: '275',
            Tags.ZERO: 'Journal Article',
            Tags.T: 'A new approach on curves of constant precession',
            Tags.A: ['Uzunoğlu, Beyhan', 'Gök, İsmail']
        }



""""Common management of citation decorators for references."""

from bibliograpy.api_bibtex import BibtexReference, default_bibtex_formatter
from bibliograpy.api_core import CitationFormatter
from bibliograpy.api_ris2001 import Tags as Ris2001, TypeFieldName as Ris2001Field, default_ris2001_formatter
from bibliograpy.api_ris2011 import Tags as Ris2011, TypeFieldName as Ris2011Field, default_ris2011_formatter
from bibliograpy.api_refer import Tags as Refer, default_refer_formatter
from bibliograpy.api_endnote import Tags as Endnote, default_endnote_formatter


class DefaultCitationFormatter(CitationFormatter):
    """Default citation formatter."""

    def __init__(self, prefix: str, itemize: str):
        self._prefix = prefix
        self._itemize = itemize

    def format(self, refs: list) -> str:

        if len(refs) == 1:
            return f"\n\n{self._prefix} {self._by_type(refs[0])}\n"

        result = f"\n\n{self._prefix}\n\n"
        for r in refs:
            result += f"{self._itemize} {self._by_type(r)}\n"
        return result

    def _by_type(self, r):
        if isinstance(r, BibtexReference):
            return self.bibtex(r)

        if isinstance(r, dict):
            if Ris2001.TY in r:
                return self.ris2001(r)
            if Ris2011.TY in r:
                return self.ris2011(r)
            if Refer.L in r:
                return self.refer(r)
            if Endnote.A in r or Endnote.T in r:
                return self.endnote(r)

        raise ValueError('unexpected reference type')

    def bibtex(self, r: BibtexReference):
        """Bibtex reference formatter."""
        return default_bibtex_formatter(r)

    def ris2001(self, r: dict[Ris2001, str | list[str] | Ris2001Field]):
        """RIS 2001 reference formatter."""
        return default_ris2001_formatter(r)

    def ris2011(self, r: dict[Ris2011, str | list[str] | Ris2011Field]):
        """RIS 2011 reference formatter."""
        return default_ris2011_formatter(r)

    def refer(self, r: dict[Refer, str | list[str]]):
        """refer reference formatter."""
        return default_refer_formatter(r)

    def endnote(self, r: dict[Endnote, str | list[str]]):
        """endnote reference formatter."""
        return default_endnote_formatter(r)

cite = DefaultCitationFormatter(prefix='Bibliography:', itemize='*')



"""Test module for RIS 2011 IO"""
from pathlib import Path
from typing import TextIO

from bibliograpy.api_core import Formats
from bibliograpy.api_ris2011 import Tags, TypeFieldName
from bibliograpy.io_ris2011 import Ris2011InputFormat


def read_ris_entries(i: TextIO):
    """ris entry parsing shortcut"""
    return Ris2011InputFormat(Formats.RIS2011).from_standard(i)

def test_multiple_records():
    """test sample using the 2011 specification"""

    with open(Path(__file__).parent / 'resources' / 'ris2011' / 'multipleRecords.ris', encoding='utf-8') as s:
        result = read_ris_entries(s)
        assert len(result) == 2
        assert result[0] == {
            Tags.TY: TypeFieldName.JOUR,
            Tags.T2: 'Bell System Technical Journal',
            Tags.PY: '1948',
            Tags.DA: 'July',
            Tags.TI: 'A Mathematical Theory of Communication',
            Tags.EP: '423',
            Tags.VL: '27',
            Tags.AU: ['Shannon, Claude E.'],
            Tags.SP: '379'
        }
        assert result[1] == {
            Tags.TY: TypeFieldName.JOUR,
            Tags.SP: '230',
            Tags.EP: '265',
            Tags.JO: 'Proc. of London Mathematical Society',
            Tags.IS: '1',
            Tags.Y1: '1937',
            Tags.A1: ['Turing, Alan Mathison'],
            Tags.T1: 'On computable numbers, with an application to the Entscheidungsproblem',
            Tags.VL: '47'
        }

def test_sample1():
    """test 2001 specification sample"""

    with open(Path(__file__).parent / 'resources' / 'ris2001' / 'sample1.ris', encoding='utf-8') as s:
        result = read_ris_entries(s)
        assert len(result) == 1
        assert result[0] == {
            Tags.TY: TypeFieldName.JOUR,
            Tags.A1: ['Baldwin,S.A.', 'Fugaccia,I.', 'Brown,D.R.', 'Brown,L.V.', 'Scheff,S.W.'],
            Tags.T1: 'Blood-brain barrier breach following cortical contusion in the rat',
            Tags.JO: 'J.Neurosurg.',
            Tags.Y1: '1996',
            Tags.VL: '85',
            Tags.SP: '476',
            Tags.EP: '481',
            Tags.RP: 'Not In File',
            Tags.KW: ['cortical contusion', 'blood-brain barrier', 'horseradish peroxidase', 'head trauma',
                      'hippocampus', 'rat'],
            Tags.N2: """Adult Fisher 344 rats were subjected to a unilateral impact to the dorsal cortex above the \
hippocampus at 3.5 m/sec with a 2 mm cortical depression. This caused severe cortical damage and neuronal loss in \
hippocampus subfields CA1, CA3 and hilus. Breakdown of the blood-brain barrier (BBB) was assessed by injecting the \
protein horseradish peroxidase (HRP) 5 minutes prior to or at various times following injury (5 minutes, 1, 2, 6, 12 \
hours, 1, 2, 5, and 10 days). Animals were killed 1 hour after HRP injection and brain sections were reacted with \
diaminobenzidine to visualize extravascular accumulation of the protein. Maximum staining occurred in animals injected \
with HRP 5 minutes prior to or 5 minutes after cortical contusion. Staining at these time points was observed in the \
ipsilateral hippocampus. Some modest staining occurred in the dorsal contralateral cortex near the superior sagittal \
sinus. Cortical HRP stain gradually decreased at increasing time intervals postinjury. By 10 days, no HRP stain was \
observed in any area of the brain. In the ipsilateral hippocampus, HRP stain was absent by 3 hours postinjury and \
remained so at the 6- and 12- hour time points. Surprisingly, HRP stain was again observed in the ipsilateral \
hippocampus 1 and 2 days following cortical contusion, indicating a biphasic opening of the BBB following head trauma \
and a possible second wave of secondary brain damage days after the contusion injury. These data indicate regions not \
initially destroyed by cortical impact, but evidencing BBB breach, may be accessible to neurotrophic factors \
administered intravenously both immediately and days after brain trauma."""
        }

def test_sample2():
    """test 2001 specification sample"""

    with open(Path(__file__).parent / 'resources' / 'ris2001' / 'sample2.ris', encoding='utf-8') as s:
        result = read_ris_entries(s)
        assert len(result) == 1
        assert result[0] == {
            Tags.TY: TypeFieldName.PAT,
            Tags.A1: ['Burger,D.R.', 'Goldstein,A.S.'],
            Tags.T1: 'Method of detecting AIDS virus infection',
            Tags.Y1: '1990/2/27',
            Tags.VL: '877609',
            Tags.IS: '4,904,581',
            Tags.RP: 'Not In File',
            Tags.A2: ['Epitope,I.'],
            Tags.CY: 'OR',
            Tags.PB: '4,629,783',
            Tags.KW: ['AIDS', 'virus', 'infection', 'antigens'],
            Tags.Y2: '1986/6/23',
            Tags.M1: 'G01N 33/569 G01N 33/577',
            Tags.M2: """435/5 424/3 424/7.1 435/7 435/29 435/32 435/70.21 435/240.27 435/172.2 530/387 530/808 530/809 \
935/110""",
            Tags.N2: """A method is disclosed for detecting the presence of HTLV III infected cells in a medium. The \
method comprises contacting the medium with monoclonal antibodies against an antigen produced as a result of the \
infection and detecting the binding of the antibodies to the antigen. The antigen may be a gene product of the HTLV \
III virus or may be bound to such gene product. On the other hand the antigen may not be a viral gene product but may \
be produced as a result of the infection and may further be bound to a lymphocyte. The medium may be a human body \
fluid or a culture medium. A particular embodiment of the present method involves a method for determining the \
presence of a AIDS virus in a person. The method comprises combining a sample of a body fluid from the person with a \
monoclonal antibody that binds to an antigen produced as a result of the infection and detecting the binding of the \
monoclonal antibody to the antigen. The presence of the binding indicates the presence of a AIDS virus infection. Also \
disclosed are novel monoclonal antibodies, noval compositions of matter, and novel diagnostic kits"""
        }

def test_sample3():
    """test 2001 specification sample"""

    with open(Path(__file__).parent / 'resources' / 'ris2001' / 'sample3.ris', encoding='utf-8') as s:
        result = read_ris_entries(s)
        assert len(result) == 1
        assert result[0] == {
            Tags.TY: TypeFieldName.CONF,
            Tags.A1: ['Catania,J.', 'Coates,T.', 'Kegeles,S.', 'Peterson,J.', 'Marin,B.', 'Fullilove,M.'],
            Tags.T1: """Predicting risk behavior with the AIDS risk reduction model (ARRM) in a random household \
probability sample of San Franciscans: the "AMEN" study""",
            Tags.Y1: '1990///6th Annual',
            Tags.VL: '6',
            Tags.SP: '318',
            Tags.EP: '318',
            Tags.RP: 'Not In File',
            Tags.CY: 'Detroit MI',
            Tags.KW: ['risk', 'AIDS', 'models', 'sexual behavior', 'HIV', 'condoms', 'heterosexual', 'bisexual',
                      'ethnicity', 'women'],
            Tags.T3: 'International Conference on AIDS 6',
            Tags.Y2: '1990/6/20',
            Tags.M1: '1',
            Tags.N1: """OBJECTIVE: Data from the AIDS In Multi-Ethnic Neighborhoods survey are used to test Stages 1 & \
3 of ARRM (a three stage process model of sexual risk behavior change; Catania, Kegeles, & Coates, 1990). Stage 1 \
analyses examine predictors of labeling one's sexual behavior in terms of HIV risk; Stage 3 concerns predictors of \
sexual behavior (e.g., condom use) (Stage 2 was not assessed in this first wave of the study but will be examined in \
wave 2). METHODS: Data were collected in a random household probability study of 1,781 white (41%), black (26%), and \
Hispanic (25%) (8% Other), unmarried respondents, aged 20-44, residing in selected \"high risk\" census tracts of San \
Francisco (Heterosexual = 83%, Homosexual = 13%, Bisexual = 4%). Labeling defined as making an accurate or inaccurate \
assessment of one's risk for HIV based on prior and current sexual practices. The behavioral outcome is frequency of \
condom use averaged across sexual partners for the past year. RESULTS: Multiple regression (Logistic & LSQ) analyses \
indicate that, 1) Accurate labeling of high risk behavior is related to high susceptibility beliefs (Imp. Chi Sq. \
=,92.46, p less than .0001), but unrelated to knowing someone with AIDS; gay relative to heterosexual men (p less than \
.03), and Hispanics compared to whites (p less than .01) were more likely to accurately label their behavior, 2) \
Greater condom use during vaginal or anal intercourse is significantly related to better sexual communication skills, \
higher perceived benefits and lower costs of condom use, but unrelated to religiosity, self-efficacy, and ethnicity \
(R's range from .50 - .66); these latter results are substantially the same for men and women, and heterosexuals and \
gay men. CONCLUSION: The findings 1) suggest the ARRM model is applicable to most social groups, 2) underscore the \
importance of interventions that enhance communication skills and teach methods of facilitating sexual enjoyment of \
condoms"""
        }

def test_sample456():
    """test 2001 specification sample"""

    with open(Path(__file__).parent / 'resources' / 'ris2001' / 'sample456.ris', encoding='utf-8') as s:
        result = read_ris_entries(s)
        assert len(result) == 3
        assert result[0] == {
            Tags.TY: TypeFieldName.RPRT,
            Tags.A1: ['Esparza,J.'],
            Tags.T1: """Report of a WHO workshop on the measurement and significance of neutralizing antibody to HIV \
and SIV, London, 3-5 October 1988""",
            Tags.Y1: '1990',
            Tags.VL: '4',
            Tags.SP: '269',
            Tags.EP: '275',
            Tags.RP: 'Not In File',
            Tags.CY: 'San Francisco CA',
            Tags.PB: 'UC Berkeley',
            Tags.KW: ['HIV', 'SIV', 'AIDS'],
            Tags.T3: 'World Health Organisation Global Programme on AIDS'
        }
        assert result[1] == {
            Tags.TY: TypeFieldName.CHAP,
            Tags.A1: ['Franks,L.M.'],
            Tags.T1: 'Preface by an AIDS Victim',
            Tags.Y1: '1991',
            Tags.VL: '10',
            Tags.SP: 'vii',
            Tags.EP: 'viii',
            Tags.RP: 'Not In File',
            Tags.T2: 'Cancer, HIV and AIDS.',
            Tags.CY: 'Berkeley CA',
            Tags.PB: 'Berkeley Press',
            Tags.KW: ['HIV', 'AIDS'],
            Tags.M1: '1',
            Tags.M2: '1',
            Tags.SN: '0-679-40110-5'
        }
        assert result[2] == {
            Tags.TY: TypeFieldName.CASE,
            Tags.A1: ['Cary,A.', 'Friedenrich,W.'],
            Tags.T1: 'Redman v. State of California',
            Tags.Y1: '1988/10/7',
            Tags.VL: '201',
            Tags.IS: '32',
            Tags.SP: '220',
            Tags.EP: '240',
            Tags.RP: 'Not In File',
            Tags.CY: 'ATLA Law Reporter',
            Tags.PB: 'San Diego County 45th Judicial District, California',
            Tags.KW: ['AIDS', 'litigation', 'AIDS litigation', 'rape'],
            Tags.U1: 'ISSN 0456-8125',
            Tags.N1: 'Raped inmate can press case against officials for contracting AIDS'
        }



from bibliograpy.api_bibtex import *

_SCOPE = dict()


BIBLIOGRAPI = Misc.generic(cite_key='bibliograpi',
                           author='Samuel Andrés',
                           title='BibliogrAPI Specification',
                           year='2024',
                           non_standard=NonStandard(url='https://cosmoloj.com/bibapi/'),
                           scope=_SCOPE)

RIS_2001 = Misc.generic(cite_key='ris_2001',
                        month='february',
                        title='RIS Format Specifications',
                        year='2001',
                        non_standard=NonStandard(url='https://web.archive.org/web/20110925024348/http://www.refman.com/support/risformat_intro.asp'),
                        scope=_SCOPE)

RIS_2011 = Misc.generic(cite_key='ris_2011',
                        month='october',
                        title='RIS Format Specifications',
                        year='2011',
                        non_standard=NonStandard(url='https://web.archive.org/web/20120616231500/http://www.refman.com/support/risformat_intro.asp'),
                        scope=_SCOPE)

REFER_MAN = Misc.generic(cite_key='refer_man',
                         title='refer(1) — Linux manual page',
                         non_standard=NonStandard(url='https://man7.org/linux/man-pages/man1/refer.1.html'),
                         scope=_SCOPE)



"""Test module for endnote process tool."""

from argparse import Namespace
from pathlib import Path
import pydoc
import sys
import time

import yaml

import pytest

from bibliograpy.api_common import cite

from bibliograpy.process import _process


def _endnote_resource(file: str) -> str:
    """Chemin vers les fichiers d'entrée."""
    return str(Path(Path(__file__).parent / 'resources' / 'endnote' / file))


def _sibbling_module(file: str) -> str:
    """Chemin vers les fichiers de modules voisins."""
    return str(Path(Path(__file__).parent / file))


def _output(file: str) -> str:
    """Chemin vers les fichiers de sortie."""
    return str(Path(Path(__file__).parent / 'resources' / 'endnote' / 'out' / file))


def test_yml_to_yml():
    """test process from a yml bibliography to a yml bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('sample1.yml'),
                       output_file=_output('test_yml_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_yml_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [
            {
                '0': 'Journal Article',
                '@': '0066-4189',
                'A': ['Le Bars, Michael', 'Cébron, David', 'Le Gal, Patrice'],
                'D': '2015',
                'I': 'Annual Reviews',
                'J': 'Annual Review of Fluid Mechanics',
                'N': '1',
                'P': '163-193',
                'T': 'Flows driven by libration, precession, and tides',
                'V': '47'
            }
        ]

def test_yml_to_endnote():
    """test process from a yml bibliography to an endnote bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('sample1.yml'),
                       output_file=_output('test_yml_to_endnote.enw'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_yml_to_endnote.enw'), encoding='utf-8') as s:
        content = s.read()
        assert content == """%0 Journal Article
%T Flows driven by libration, precession, and tides
%A Le Bars, Michael
%A Cébron, David
%A Le Gal, Patrice
%J Annual Review of Fluid Mechanics
%V 47
%N 1
%P 163-193
%@ 0066-4189
%D 2015
%I Annual Reviews

"""

def test_yml_to_json():
    """test process from a yml bibliography to a json bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('sample1.yml'),
                       output_file=_output('test_yml_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_yml_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"0": "Journal Article", '
                           '"T": "Flows driven by libration, precession, and tides", '
                           '"A": ["Le Bars, Michael", "C\\u00e9bron, David", "Le Gal, Patrice"], '
                           '"J": "Annual Review of Fluid Mechanics", '
                           '"V": "47", '
                           '"N": "1", '
                           '"P": "163-193", '
                           '"@": "0066-4189", '
                           '"D": "2015", '
                           '"I": "Annual Reviews"}]')

def test_json_to_yml():
    """test process from a json bibliography to a yml bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('sample1.json'),
                       output_file=_output('test_json_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_json_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [
            {
                '0': 'Journal Article',
                '@': '0066-4189',
                'A': ['Le Bars, Michael', 'Cébron, David', 'Le Gal, Patrice'],
                'D': '2015',
                'I': 'Annual Reviews',
                'J': 'Annual Review of Fluid Mechanics',
                'N': '1',
                'P': '163-193',
                'T': 'Flows driven by libration, precession, and tides',
                'V': '47'
            }
        ]

def test_json_to_endnote():
    """test process from a json bibliography to an endnote bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('sample1.json'),
                       output_file=_output('test_json_to_endnote.enw'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_json_to_endnote.enw'), encoding='utf-8') as s:
        content = s.read()
        assert content == """%0 Journal Article
%T Flows driven by libration, precession, and tides
%A Le Bars, Michael
%A Cébron, David
%A Le Gal, Patrice
%J Annual Review of Fluid Mechanics
%V 47
%N 1
%P 163-193
%@ 0066-4189
%D 2015
%I Annual Reviews

"""

def test_json_to_json():
    """test process from a json bibliography to a json bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('sample1.json'),
                       output_file=_output('test_json_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_json_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"0": "Journal Article", '
                           '"T": "Flows driven by libration, precession, and tides", '
                           '"A": ["Le Bars, Michael", "C\\u00e9bron, David", "Le Gal, Patrice"], '
                           '"J": "Annual Review of Fluid Mechanics", '
                           '"V": "47", '
                           '"N": "1", '
                           '"P": "163-193", '
                           '"@": "0066-4189", '
                           '"D": "2015", '
                           '"I": "Annual Reviews"}]')

def test_endnote_to_yml():
    """test process from an endnote bibliography to a yml bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('sample1.enw'),
                       output_file=_output('test_endnote_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_endnote_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [
            {
                '0': 'Journal Article',
                '@': '0066-4189',
                'A': ['Le Bars, Michael', 'Cébron, David', 'Le Gal, Patrice'],
                'D': '2015',
                'I': 'Annual Reviews',
                'J': 'Annual Review of Fluid Mechanics',
                'N': '1',
                'P': '163-193',
                'T': 'Flows driven by libration, precession, and tides',
                'V': '47'
            }
        ]

def test_endnote_to_endnote():
    """test process from an endnote bibliography to an endnote bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('sample1.enw'),
                       output_file=_output('test_endnote_to_endnote.enw'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_endnote_to_endnote.enw'), encoding='utf-8') as s:
        content = s.read()
        assert content == """%0 Journal Article
%T Flows driven by libration, precession, and tides
%A Le Bars, Michael
%A Cébron, David
%A Le Gal, Patrice
%J Annual Review of Fluid Mechanics
%V 47
%N 1
%P 163-193
%@ 0066-4189
%D 2015
%I Annual Reviews

"""

def test_endnote_to_json():
    """test process from an endnote bibliography to a json bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('sample1.enw'),
                       output_file=_output('test_endnote_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_endnote_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"0": "Journal Article", '
                           '"T": "Flows driven by libration, precession, and tides", '
                           '"A": ["Le Bars, Michael", "C\\u00e9bron, David", "Le Gal, Patrice"], '
                           '"J": "Annual Review of Fluid Mechanics", '
                           '"V": "47", '
                           '"N": "1", '
                           '"P": "163-193", '
                           '"@": "0066-4189", '
                           '"D": "2015", '
                           '"I": "Annual Reviews"}]')

def test_yml_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('mini.yml'),
                       output_file=_sibbling_module('test_endnote_yml_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_endnote_yml_to_py import INTERNATIONAL_ASTRONOMICAL_UNION, NASA

    @cite(INTERNATIONAL_ASTRONOMICAL_UNION, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_endnote

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union
    * NASA
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_endnote

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union
    * NASA
""")

def test_json_to_py():
    """test process from a json bibliography to a py source bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('mini.json'),
                       output_file=_sibbling_module('test_endnote_json_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_endnote_json_to_py import INTERNATIONAL_ASTRONOMICAL_UNION, NASA

    @cite(INTERNATIONAL_ASTRONOMICAL_UNION, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_endnote

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union
    * NASA
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_endnote

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union
    * NASA
""")

def test_endnote_to_py():
    """test process from an endnote bibliography to a py source bibliography"""

    _process(Namespace(CMD='endnote',
                       file=_endnote_resource('mini.enw'),
                       output_file=_sibbling_module('test_endnote_endnote_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_endnote_endnote_to_py import INTERNATIONAL_ASTRONOMICAL_UNION, NASA

    @cite(INTERNATIONAL_ASTRONOMICAL_UNION, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_endnote

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union
    * NASA
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_endnote

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union
    * NASA
""")

def test_input_file_not_found():
    """test process input file not found"""

    with pytest.raises(FileNotFoundError) as e:
        with open(_endnote_resource('not_existing_file.yml'), encoding='utf-8') as s:
            yaml.safe_load(s)

    assert e.value.args[0] == 2
    assert e.value.args[1] == "No such file or directory"



"""Core management of reference decorators."""
from dataclasses import dataclass
from enum import Enum
from typing import TextIO

@dataclass(frozen=True)
class Format:
    """A format representation."""
    title: str | None
    command: str | None
    io_extension: list[str]

class Formats(Format, Enum):
    """Supported bibliography formats."""
    BIBTEX = ('Bibtex', 'bibtex', ['bib', 'bibtex'])
    RIS2001 = ('RIS (2001)', 'ris2001', ['ris'])
    RIS2011 = ('RIS (2011)', 'ris2011', ['ris'])
    REFER = ('refer', 'refer', ['refer'])
    ENDNOTE = ('Endnote', 'endnote', ['enw'])
    YML = (None, None, ['yml', 'yaml'])
    JSON = (None, None, ['json'])
    PYTHON = (None, None, ['py'])

    @staticmethod
    def as_command(format_id: str):
        """Gets a supported format enum instance from a supported process argument string."""
        for f in Formats:
            if format_id == f.command:
                return f
        raise ValueError(f'unexpected format {format_id}')

    @staticmethod
    def as_io_extension(format_id: str):
        """Gets a supported format enum instance from a supported process argument string."""
        for f in Formats:
            if format_id in f.io_extension:
                return f
        raise ValueError(f'unexpected format {format_id}')


class InputFormat:
    """InputFormat interface to deserialize a bibliography."""

    def __init__(self, source: Format, standard: Format):
        self._source = source
        self._standard = standard

    def from_yml(self, i: TextIO):
        """Reads from yml representation."""

    def from_json(self, i: TextIO):
        """Reads from json representation."""

    def from_standard(self, i: TextIO):
        """Reads from standard format."""

    def source(self) -> Format:
        """The source format."""
        return self._source

    def standard(self) -> Format:
        """The standard format."""
        return self._standard

    def read(self, i: TextIO):
        """Deserialization method."""
        if self.source() is Formats.YML:
            return self.from_yml(i)
        if self.source() is Formats.JSON:
            return self.from_json(i)
        if self.source() is self.standard():
            return self.from_standard(i)

        raise ValueError(f'unsupported configuration format {self.source()}')

class OutputFormat:
    """Output format to serialize a bibliography."""

    def __init__(self, target: Format, standard: Format):
        self._target = target
        self._standard = standard

    def to_yml(self, o: TextIO):
        """Writes to yml representation."""

    def to_json(self, o: TextIO):
        """Writes to json representation."""

    def to_standard(self, o: TextIO):
        """Writes to standard format."""

    def to_py(self, o: TextIO):
        """Writes to python representation."""

    def target(self):
        """The file extension."""
        return self._target

    def standard(self) -> Format:
        """The standard format."""
        return self._standard

    def write(self, o: TextIO):
        """Serialization method."""
        if self.target() is Formats.YML:
            return self.to_yml(o)
        if self.target() is Formats.JSON:
            return self.to_json(o)
        if self.target() is Formats.PYTHON:
            return self.to_py(o)
        if self.target() is self.standard():
            return self.to_standard(o)

        raise ValueError(f'unsupported configuration format {self.target()}')


class CitationFormatter:
    """A builder of reference decorators."""

    def format(self, refs: list) -> str:
        """Formats a citation list."""

    def __call__(self, *refs):
        """The reference decorator."""

        def internal(obj):
            if obj.__doc__ is None:
                obj.__doc__ = ''
            if len(refs) == 1:
                ref0 = refs[0]
                if isinstance(ref0, list):
                    obj.__doc__ += self.format(ref0)
                else:
                    obj.__doc__ += self.format([ref0])
            else:
                obj.__doc__ += self.format([*refs])
            return obj

        return internal


class SimpleCitationFormatter(CitationFormatter):
    """A simple citation formatter for """

    def __init__(self, prefix, itemize, reference_formatter):
        self._prefix = prefix
        self._itemize = itemize
        self._reference_formatter = reference_formatter

    def format(self, refs: list) -> str:

        if len(refs) == 1:
            return f"\n\n{self._prefix} {self._reference_formatter(refs[0])}\n"

        result = f"\n\n{self._prefix}\n\n"
        for r in refs:
            result += f"{self._itemize} {self._reference_formatter(r)}\n"
        return result



"""refer I/O module."""

import json

from typing import TextIO

import yaml

from bibliograpy.api_core import InputFormat, OutputFormat, Format, Formats
from bibliograpy.api_refer import Tags

class ReferInputFormat(InputFormat):
    """refer input format implementation."""

    def __init__(self, source: Format):
        super().__init__(source=source, standard=Formats.REFER)

    def from_yml(self, i: TextIO):
        """Reads from yml representation."""
        return [{Tags.parse(k): e[k] for k in e} for e in yaml.safe_load(i)]

    def from_json(self, i: TextIO):
        """Reads from json representation."""
        return [{Tags.parse(k): e[k] for k in e} for e in json.load(i)]

    def from_standard(self, i: TextIO) -> list[dict[Tags, str | list[str]]]:
        """Reads from standard format."""

        results: list[dict[Tags, str | list[str]]] = []

        result: dict[Tags, str | list[str]] = {}
        while line := i.readline():
            if line.strip() == '':
                if result:
                    results.append(result)
                    result = {}
                continue

            tag = Tags.parse(line[:2])

            if tag.repeating:
                if tag in result:
                    result[tag].append(line[3:].rstrip('\n\r'))
                else:
                    result[tag] = [line[3:].rstrip('\n\r')]
            else:
                result[tag] = line[3:].rstrip('\n\r')

        if result:
            results.append(result)
        return results


class ReferOutputFormat(OutputFormat):
    """refer format implementation."""

    def __init__(self,
                 content: list[dict[Tags, str | list[str]]],
                 target: Format):
        super().__init__(target=target, standard=Formats.REFER)
        self._content = content

    def to_yml(self, o: TextIO):
        """Writes to yml representation."""
        yaml.dump([{k.name: e[k] for k in e} for e in self._content],
                  o,
                  sort_keys=False)

    def to_json(self, o: TextIO):
        """Writes to json representation."""
        json.dump([{k.name: e[k] for k in e} for e in self._content],
                  fp=o,
                  sort_keys=False)

    def to_standard(self, o: TextIO):
        """Writes to standard format."""

        for bib_entry in self._content:

            for tag in bib_entry:

                if tag.repeating:
                    for l in bib_entry[tag]:
                        o.write(f'%{tag.name} {l}\n')
                else:
                    o.write(f'%{tag.name} {bib_entry[tag]}\n')

            o.write('\n')

    def to_py(self, o: TextIO):
        """Writes to python representation."""

        o.write('from bibliograpy.api_refer import *\n\n')

        for bib_entry in self._content:
            o.write(f'{bib_entry[Tags.L].upper()} = ')
            o.write('{\n')
            for e in bib_entry:
                o.write(f"  Tags.{e.name}: '{bib_entry[e]}',\n")
            o.write('}\n')



"""Test module for refer process tool."""

from argparse import Namespace
from pathlib import Path
import pydoc
import sys
import time

import yaml

import pytest

from bibliograpy.api_common import cite

from bibliograpy.process import _process


def _refer_resource(file: str) -> str:
    """Chemin vers les fichiers d'entrée."""
    return str(Path(Path(__file__).parent / 'resources' / 'refer' / file))


def _sibbling_module(file: str) -> str:
    """Chemin vers les fichiers de modules voisins."""
    return str(Path(Path(__file__).parent / file))


def _output(file: str) -> str:
    """Chemin vers les fichiers de sortie."""
    return str(Path(Path(__file__).parent / 'resources' / 'refer' / 'out' / file))


def test_yml_to_yml():
    """test process from a yml bibliography to a yml bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('sample1.yml'),
                       output_file=_output('test_yml_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_yml_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [
            {
                'A': ['Brian W. Kernighan', 'Lorinda L. Cherry'],
                'T': 'A System for Typesetting Mathematics',
                'J': 'J. Comm. ACM',
                'V': '18',
                'N': '3',
                'D': 'March 1978',
                'P': '151-157',
                'K': 'eqn'
            }]

def test_yml_to_refer():
    """test process from a yml bibliography to a RIS 2001 bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('sample1.yml'),
                       output_file=_output('test_yml_to_refer.refer'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_yml_to_refer.refer'), encoding='utf-8') as s:
        content = s.read()
        assert content == """%A Brian W. Kernighan
%A Lorinda L. Cherry
%T A System for Typesetting Mathematics
%J J. Comm. ACM
%V 18
%N 3
%D March 1978
%P 151-157
%K eqn

"""

def test_yml_to_json():
    """test process from a yml bibliography to a RIS 2001 bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('sample1.yml'),
                       output_file=_output('test_yml_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_yml_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"A": ["Brian W. Kernighan", "Lorinda L. Cherry"], '
                           '"T": "A System for Typesetting Mathematics", '
                           '"J": "J. Comm. ACM", '
                           '"V": "18", '
                           '"N": "3", '
                           '"D": "March 1978", '
                           '"P": "151-157", '
                           '"K": "eqn"}]')

def test_json_to_yml():
    """test process from a json bibliography to a yml bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('sample1.json'),
                       output_file=_output('test_json_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_json_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [
            {
                'A': ['Brian W. Kernighan', 'Lorinda L. Cherry'],
                'T': 'A System for Typesetting Mathematics',
                'J': 'J. Comm. ACM',
                'V': '18',
                'N': '3',
                'D': 'March 1978',
                'P': '151-157',
                'K': 'eqn'
            }]

def test_json_to_refer():
    """test process from a json bibliography to a RIS 2001 bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('sample1.json'),
                       output_file=_output('test_json_to_refer.refer'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_json_to_refer.refer'), encoding='utf-8') as s:
        content = s.read()
        assert content == """%A Brian W. Kernighan
%A Lorinda L. Cherry
%T A System for Typesetting Mathematics
%J J. Comm. ACM
%V 18
%N 3
%D March 1978
%P 151-157
%K eqn

"""

def test_json_to_json():
    """test process from a json bibliography to a json bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('sample1.json'),
                       output_file=_output('test_json_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_json_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"A": ["Brian W. Kernighan", "Lorinda L. Cherry"], '
                           '"T": "A System for Typesetting Mathematics", '
                           '"J": "J. Comm. ACM", '
                           '"V": "18", '
                           '"N": "3", '
                           '"D": "March 1978", '
                           '"P": "151-157", '
                           '"K": "eqn"}]')

def test_refer_to_yml():
    """test process from a bib bibliography to a yml bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('sample1.refer'),
                       output_file=_output('test_refer_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_refer_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [
            {
                'A': ['Brian W. Kernighan', 'Lorinda L. Cherry'],
                'T': 'A System for Typesetting Mathematics',
                'J': 'J. Comm. ACM',
                'V': '18',
                'N': '3',
                'D': 'March 1978',
                'P': '151-157',
                'K': 'eqn'
            }]

def test_refer_to_refer():
    """test process from a RIS 2001 bibliography to a RIS 2001 bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('sample1.refer'),
                       output_file=_output('test_refer_to_refer.refer'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_refer_to_refer.refer'), encoding='utf-8') as s:
        content = s.read()
        assert content == """%A Brian W. Kernighan
%A Lorinda L. Cherry
%T A System for Typesetting Mathematics
%J J. Comm. ACM
%V 18
%N 3
%D March 1978
%P 151-157
%K eqn

"""

def test_refer_to_json():
    """test process from a RIS 2001 bibliography to a json bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('sample1.refer'),
                       output_file=_output('test_refer_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_output('test_refer_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"A": ["Brian W. Kernighan", "Lorinda L. Cherry"], '
                           '"T": "A System for Typesetting Mathematics", '
                           '"J": "J. Comm. ACM", '
                           '"V": "18", '
                           '"N": "3", '
                           '"D": "March 1978", '
                           '"P": "151-157", '
                           '"K": "eqn"}]')

def test_yml_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('mini.yml'),
                       output_file=_sibbling_module('test_refer_yml_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_refer_yml_to_py import IAU, NASA

    @cite(IAU, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_refer

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union [iau]
    * NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_refer

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union [iau]
    * NASA [nasa]
""")

def test_json_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('mini.json'),
                       output_file=_sibbling_module('test_refer_json_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_refer_json_to_py import IAU, NASA

    @cite(IAU, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_refer

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union [iau]
    * NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_refer

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union [iau]
    * NASA [nasa]
""")

def test_refer_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='refer',
                       file=_refer_resource('mini.refer'),
                       output_file=_sibbling_module('test_refer_refer_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_refer_refer_to_py import IAU, NASA

    @cite(IAU, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_refer

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union [iau]
    * NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_process_refer

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union [iau]
    * NASA [nasa]
""")

def test_input_file_not_found():
    """test process input file not found"""

    with pytest.raises(FileNotFoundError) as e:
        with open(_refer_resource('not_existing_file.yml'), encoding='utf-8') as s:
            yaml.safe_load(s)

    assert e.value.args[0] == 2
    assert e.value.args[1] == "No such file or directory"



"""Bibtex I/O module."""

import json

from typing import Any, TextIO

import bibtexparser
import yaml
from bibtexparser.bibdatabase import BibDatabase
from bibtexparser.bwriter import BibTexWriter

from bibliograpy.api_bibtex import TYPES, BibtexReference, ENTRYTYPE_FIELD_IN_MODEL_DICT, ID_FIELD_IN_MODEL_DICT
from bibliograpy.api_core import InputFormat, OutputFormat, Formats, Format

_ENTRY_TYPE_FIELD_IN_DICT = 'entry_type'


class BibtexInputFormat(InputFormat):
    """Bibtex input format implementation."""

    def __init__(self, source: Format):
        super().__init__(source=source, standard=Formats.BIBTEX)

    def from_yml(self, i: TextIO):
        """Reads from yml representation."""
        return yaml.safe_load(i)

    def from_json(self, i: TextIO):
        """Reads from json representation."""
        return json.load(i)

    def from_standard(self, i: TextIO):
        """Reads from standard format."""
        meta = {}
        content = []
        for e in bibtexparser.load(i).entries:
            meta[_ENTRY_TYPE_FIELD_IN_DICT] = e[ENTRYTYPE_FIELD_IN_MODEL_DICT]
            meta[BibtexReference.CITE_KEY_FIELD] = e[ID_FIELD_IN_MODEL_DICT]
            del e[ENTRYTYPE_FIELD_IN_MODEL_DICT]
            del e[ID_FIELD_IN_MODEL_DICT]
            content.append({**meta, **e})
        return content

class BibtexOutputFormat(OutputFormat):
    """Bibtex format implementation."""

    def __init__(self,
                 content: list[dict],
                 target: Format,
                 scope_symbol: str | None,
                 init_scope: str | None):
        super().__init__(target=target, standard=Formats.BIBTEX)
        self._content = content
        self._scope_symbol = scope_symbol
        self._init_scope = init_scope

    def _to_standard_ext(self) -> list[str]:
        """Extensions to be serialized to standard format."""
        return ['bib', 'bibtex']

    def to_yml(self, o: TextIO):
        """Writes to yml representation."""
        yaml.dump(self._content, o, sort_keys=False)

    def to_json(self, o: TextIO):
        """Writes to json representation."""
        json.dump(self._content, fp=o, sort_keys=False)

    def to_standard(self, o: TextIO):
        """Writes to standard format."""

        scope: dict[str, Any] = {}
        entries = []

        for ref in self._content:
            entry_type = ref[_ENTRY_TYPE_FIELD_IN_DICT]
            if entry_type in TYPES:
                entries.append(TYPES[entry_type].from_dict(ref, scope).to_bib())

        db = BibDatabase()
        db.entries = entries
        writer = BibTexWriter()
        writer.order_entries_by = None

        bibtexparser.dump(bib_database=db, bibtex_file=o, writer=writer)

    def to_py(self, o: TextIO):
        """Writes to python representation."""
        scope: dict[str, Any] = {}

        o.write('from bibliograpy.api_bibtex import *\n')
        o.write('\n')

        if self._init_scope is not None:
            o.write(f'{self._scope_symbol} = {self._init_scope}\n')
            o.write('\n')

        for ref in self._content:
            entry_type = ref[_ENTRY_TYPE_FIELD_IN_DICT]
            if entry_type in TYPES:
                o.write(f"{TYPES[entry_type].from_dict(ref, scope).to_py(scope_symbol=self._scope_symbol)}\n")



"""Test module for core api"""
import sys
import pydoc

from bibliograpy.api_bibtex import Misc, TechReport, BibtexReference
from bibliograpy.api_core import CitationFormatter

SCOPE: dict[str, BibtexReference] = {}


IAU = Misc.generic(cite_key='iau',
                   title='International Astronomical Union',
                   institution='IAU',
                   scope=SCOPE)


IAU_2006_B1 = TechReport.generic(
    cite_key='iau_2006_b1',
    author='',
    crossref='iau',
    title='Adoption of the P03 Precession Theory and Definition of the Ecliptic',
    year=2006,
    scope=SCOPE)


def test_custom_citation_formatter():
    """test custom citation formatter"""

    class _CitationFormatter(CitationFormatter):
        def format(self, refs: list) -> str:
            if len(refs) == 1:
                return f"\n\nBibliographie : {self.ref_formatter(refs[0])}\n"

            result = "\n\nBibliographie :\n\n"
            for r in refs:
                result += f"* {self.ref_formatter(r)}\n"
            return result

        def ref_formatter(self, r: BibtexReference) -> str:
            """The single reference formatter."""
            return f"{r.title} [{r.cite_key}]"

    ref = _CitationFormatter()

    @ref(IAU_2006_B1, IAU)
    def tatafr():
        """ma doc avec plusieurs références en varargs"""


    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(tatafr) ==
"""Python Library Documentation: function tatafr in module test_api_core

t\bta\bat\bta\baf\bfr\br()
    ma doc avec plusieurs références en varargs

    Bibliographie :

    * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
    * International Astronomical Union [iau]
""")
    else:
        assert (pydoc.render_doc(tatafr) ==
"""Python Library Documentation: function tatafr in module test_api_core

t\bta\bat\bta\baf\bfr\br()
    ma doc avec plusieurs références en varargs
    
    Bibliographie :
    
    * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
    * International Astronomical Union [iau]
""")



"""
RIS 2011 specification model.
"""
from dataclasses import dataclass
from enum import Enum, auto

from bibliograpy.bibliography import RIS_2011
from bibliograpy.api_bibtex import _cite


@dataclass(frozen=True)
class Tag:
    """A field tag."""
    auto: auto
    repeating: bool = False

@_cite(RIS_2011)
class Tags(Tag, Enum):
    """
    RIS fields.
    """

    TY = auto()
    """Type of reference. 
    This must contain the abbreviation for the reference type as found in the next section. This will determine how all 
    other fields are interpreted."""

    ER = auto()
    """End of reference.
    Must be the last tag in a reference."""

    AU = (auto(), True)
    """Authors, Editors, Translators. 
    Each author must be on a separate line, preceded by the tag that corresponds to the author role (see individual ref 
    type matrix for role definitions). Each reference can contain unlimited author fields, and can contain up to 255 
    characters for each field. The author name must be in the following syntax:
    
    Lastname, Firstname, Suffix
    
    For Firstname, you can use full names, initials, or both. The format for the author’s first name is as follows:
    
    Phillips, A.J
    
    Phillips, Albert John
    
    Phillips, Albert
    
    Lastname = Any string of letters, spaces, and hyphens
    
    Firstname = Any string of letters, spaces, and hyphens
    
    Initial = Any single letter followed by a period
    
    Full Name = Any string of letters, spaces, and hyphens
    
    Suffix = Jr/Sr/II/III/MD etc. (Phillips,A.J.,Sr.); use of the suffix is optional"""

    A2 = (auto(), True)
    """Secondary Author"""

    A3 = (auto(), True)
    """Tertiary Author"""

    A4 = (auto(), True)
    """Subsidiary Author"""

    AB = auto()
    """Abstract"""

    AD = auto()
    """Author address"""

    AN = auto()
    """Accession Number"""

    C1 = auto()
    """Custom 1"""

    C2 = auto()
    """Custom 2"""

    C3 = auto()
    """Custom 3"""

    C4 = auto()
    """Custom 4"""

    C5 = auto()
    """Custom 5"""

    C6 = auto()
    """Custom 6"""

    C7 = auto()
    """Custom 7"""

    C8 = auto()
    """Custom 8"""

    CA = auto()
    """Caption"""

    CN = auto()
    """Call Number"""

    CY = auto()
    """Place Published"""

    DB = auto()
    """Name of Database"""

    DO = auto()
    """DOI"""

    DP = auto()
    """Database provider"""

    ET = auto()
    """Edition"""

    ID = auto()
    """The characters allowed in the reference ID fields can be in the set "0" through "9," or "A" through "Z.\""""

    PY = auto()
    """This is the publication year. 
    It must be four numeric characters. Dates prior to 1000 should use “0” in the positions before the intended date. 
    For example, the year 765 would be represented as 0765."""

    DA = auto()
    """Dates must be in the following format:
    
    YYYY/MM/DD/other info
    
    The year, month and day fields are all numeric. The other info field can be any string of letters, spaces and 
    hyphens.
    
    Note that each specific date information is optional, however the slashes (“/”) are not. For example, if you just 
    had the <year> and <other info>, then the output would look like: “1993///Spring.”
    
    Date information should be set forth in the following format:
    
    YYYY or
    YYYY/MM or
    YYYY/MM/DD or
    YYYY/MM/DD/other info"""

    KW = (auto(), True)
    """Keywords.
    Each keyword or phrase must be on its own line, preceded by this tag. A keyword can consist of multiple words 
    (phrases) and can be up to 255 characters long. There can be unlimited keywords in a reference."""

    RP = auto()
    """Reprint status.
    This optional field can contain one of three status notes. Each must be in uppercase, and the date after 
    “ON REQUEST” must be in USA format, in parentheses (MM/DD/YY). If this field is blank the Import function assumes 
    the reprint status is “NOT IN FILE.”
    
    The three options are:

    IN FILE – The data provider has a corresponding physical copy for the reference.
    NOT IN FILE – The data provider does not have a corresponding physical copy.
    ON REQUEST (mm/dd/yy) - This means that the data provider has requested a reprint of the reference; the date is the 
    date on which the reprint was requested (in mm/dd/yy format)."""

    J2 = auto()
    """Periodical name: standard abbreviation. 
    This is the abbreviation of the periodical in which the article is published. If possible, periodical names should 
    be abbreviated in the Index Medicus® style, with periods after the abbreviations. This field is mapped to the full 
    journal name in T2 and is used as the journal abbreviation in output styles."""

    L1 = auto()
    """File attachment"""

    L4 = auto()
    """Figure"""

    LA = auto()
    """Language"""

    LB = auto()
    """Label"""

    IS = auto()
    """Number"""

    M3 = auto()
    """Type of Work"""

    N1 = auto()
    """Notes"""

    NV = auto()
    """Number of Volumes"""

    OP = auto()
    """Original Publication"""

    PB = auto()
    """Publisher"""

    UR = auto()
    """Web/URL.
    There is no practical length limit to this field. URL addresses can be entered individually, one per tag, or
    multiple addresses can be entered on one line using a semi-colon as a separator."""

    # following fields are for implicit RIS 2001 retrocompatibility in RIS 2011 samples
    T3 = auto()
    CT = auto()
    U3 = auto()
    L3 = auto()
    T1 = auto()
    CP = auto()
    AV = auto()
    EP = auto()
    JO = auto()
    U1 = auto()
    U5 = auto()
    VL = auto()
    L2 = auto()
    JF = auto()
    U4 = auto()
    ED = (auto(), True)
    J1 = auto()
    SN = auto()
    Y1 = auto()
    T2 = auto()
    N2 = auto()
    TI = auto()
    BT = auto()
    A1 = (auto(), True)
    JA = auto()
    M2 = auto()
    U2 = auto()
    M1 = auto()
    Y2 = auto()
    SP = auto()

    @staticmethod
    def parse(tag_str: str):
        """Parses a tag name into an enum value."""
        for n in Tags:
            if tag_str == n.name:
                return n
        raise ValueError(f'unknown {tag_str} tag')


@_cite(RIS_2011)
class TypeFieldName(Enum):
    """Reference Type Field Names

    The following describes the valid reference type field names that can be used with for the reference type field when
    importing references into Reference Manager."""

    ABST = auto()
    """Abstract"""

    AGGR = auto()
    """Aggregated Database"""

    ANCIENT = auto()
    """Ancient Text"""

    ADVS = auto()
    """Audiovisual material"""

    ART = auto()
    """Art Work"""

    BILL = auto()
    """Bill"""

    BLOG = auto()
    """Blog"""

    BOOK = auto()
    """Book, Whole"""

    CASE = auto()
    """Case"""

    CHAP = auto()
    """Book Section"""

    CHART = auto()
    """Chart"""

    CLSWK = auto()
    """Classical Work"""

    COMP = auto()
    """Computer program"""

    CONF = auto()
    """Conference proceeding"""

    CPAPER = auto()
    """Conference Paper"""

    CTLG = auto()
    """Catalog"""

    DATA = auto()
    """Dataset"""

    DICT = auto()
    """Dictionary"""

    EDBOOK = auto()
    """Edited book"""

    EBOOK = auto()
    """Electronic book"""

    ECHAP = auto()
    """Electronic Book Section"""

    EJOUR = auto()
    """Electronic Article"""

    ENCYC = auto()
    """Encyclopedia"""

    EQUA = auto()
    """Equation"""

    FIGURE = auto()
    """Figure"""

    GEN = auto()
    """Generic"""

    GOVDOC = auto()
    """Government Document"""

    GRNT = auto()
    """Grant"""

    HEAR = auto()
    """Hearing"""

    ICOMM = auto()
    """Internet Communication"""

    INPR = auto()
    """In Press Article"""

    JFULL = auto()
    """Full Journal"""

    JOUR = auto()
    """Journal"""

    LEGAL = auto()
    """Legal Rule"""

    MAP = auto()
    """Map"""

    MGZN = auto()
    """Magazine article"""

    MANSCPT = auto()
    """Manuscript"""

    MUSIC = auto()
    """Music"""

    NEWS = auto()
    """Newspaper"""

    DBASE = auto()
    """Online Database"""

    MULTI = auto()
    """Online Multimedia"""

    PAMP = auto()
    """Pamphlet"""

    PAT = auto()
    """Patent"""

    PCOMM = auto()
    """Personal communication"""

    RPRT = auto()
    """Report"""

    SER = auto()
    """Serial (Book, Monograph)"""

    SLIDE = auto()
    """Slide"""

    SOUND = auto()
    """Sound recording"""

    STAND = auto()
    """Standard"""

    STAT = auto()
    """Statute"""

    THES = auto()
    """Thesis/Dissertation"""

    UNBILL = auto()
    """Unenacted bill/resolution"""

    UNPD = auto()
    """Unpublished work"""

    VIDEO = auto()
    """Video recording"""

    ELEC = auto()
    """Web Page"""

    @staticmethod
    def parse(entry_type: str):
        """Parses an entry type name into an enum value."""
        for n in TypeFieldName:
            if entry_type == n.name:
                return n
        raise ValueError(f'unknown {entry_type} type')

def default_ris2011_formatter(r: dict[Tags, str | list[str] | TypeFieldName]):
    """The default formatter for RIS 2011 references."""
    title = r[Tags.TI] if Tags.TI in r else (r[Tags.T1] if Tags.T1 in r else (r[Tags.CT] if Tags.CT in r else ""))
    return f"{title} [{r[Tags.ID]}]" if Tags.ID in r else title



"""RIS I/O module."""

import json

from typing import TextIO

import yaml

from bibliograpy.api_core import InputFormat, OutputFormat, Format, Formats
from bibliograpy.api_ris2001 import Tags, TypeFieldName

class Ris2001InputFormat(InputFormat):
    """Ris 2001 input format implementation."""

    def __init__(self, source: Format):
        super().__init__(source=source, standard=Formats.RIS2001)

    def from_yml(self, i: TextIO):
        """Reads from yml representation."""
        return [{Tags.parse(k): TypeFieldName.parse(e[k]) if Tags.parse(k) is Tags.TY else e[k] for k in e}
                for e in yaml.safe_load(i)]

    def from_json(self, i: TextIO):
        """Reads from json representation."""
        return [{Tags.parse(k): TypeFieldName.parse(e[k]) if Tags.parse(k) is Tags.TY else e[k] for k in e}
                for e in json.load(i)]

    def from_standard(self, i: TextIO) -> list[dict[Tags, str | list[str] | TypeFieldName]]:
        """Reads from standard format."""

        results: list[dict[Tags, str | list[str] | TypeFieldName]] = []

        while line := i.readline():
            if line.rstrip() == '':
                continue
            entry: dict[Tags, str | list[str] | TypeFieldName] = {Tags.TY: _parse_ris_entry_type(line)}
            entry.update(_read_ris_entry(i))
            results.append(entry)
        return results

def _parse_ris_entry_type(line: str) -> TypeFieldName:
    # first field must contain entry type
    tag = Tags.parse(line[:2])
    if tag is not Tags.TY:
        raise ValueError(f'expected type field but found {tag}')

    if line[2:6] != '  - ':
        raise ValueError(f'type line "{line}" is not correctly formatted')

    return TypeFieldName.parse(line[6:].rstrip())

def _read_ris_entry(tio: TextIO) -> dict[Tags, str | list[str]]:
    """Reads a single RIS entry from the input stream."""

    result = {}

    last_tag: Tags | None = None

    while line := tio.readline():

        try:
            tag = Tags.parse(line[:2])
            last_tag = tag

            if tag is Tags.ER:
                return result

            if tag is Tags.TY:
                raise ValueError('only one type field is expected, a ')

            if tag.repeating:
                if tag in result:
                    result[tag].append(line[6:].rstrip('\n\r'))
                else:
                    result[tag] = [line[6:].rstrip('\n\r')]
            else:
                result[tag] = line[6:].rstrip('\n\r')
        except ValueError as e:
            if line[2:6] == '  - ' or last_tag is None:
                raise e

            # long field support
            if last_tag.repeating:
                result[last_tag][-1] += line.rstrip('\n\r')
            else:
                result[last_tag] += line.rstrip('\n\r')
    raise ValueError(f'the last RIS entry tag is expected to be {Tags.ER.name} but found {last_tag}')


class Ris2001OutputFormat(OutputFormat):
    """Bibtex format implementation."""

    def __init__(self,
                 content: list[dict[Tags, str | list[str] | TypeFieldName]],
                 target: Format):
        super().__init__(target=target, standard=Formats.RIS2001)
        self._content = content

    def to_yml(self, o: TextIO):
        """Writes to yml representation."""
        yaml.dump([{k.name: (e[k].name if isinstance(e[k], TypeFieldName) else e[k]) for k in e}
                   for e in self._content],
                  o,
                  sort_keys=False)

    def to_json(self, o: TextIO):
        """Writes to json representation."""
        json.dump([{k.name: (e[k].name if isinstance(e[k], TypeFieldName) else e[k]) for k in e}
                   for e in self._content],
                  fp=o,
                  sort_keys=False)

    def to_standard(self, o: TextIO):
        """Writes to standard format."""

        for bib_entry in self._content:
            o.write(f'{Tags.TY.name}  - {bib_entry[Tags.TY].name}')
            o.write('\n')

            for tag in bib_entry:

                if tag is Tags.TY:
                    continue

                if tag.repeating:
                    for l in bib_entry[tag]:
                        o.write(f'{tag.name}  - {l}\n')
                else:
                    o.write(f'{tag.name}  - {bib_entry[tag]}\n')

            o.write(f'{Tags.ER.name}  - \n')

    def to_py(self, o: TextIO):
        """Writes to python representation."""

        o.write('from bibliograpy.api_ris2001 import *\n\n')
        o.write('\n')

        for bib_entry in self._content:
            o.write(f'{bib_entry[Tags.ID].upper()} = ')
            o.write('{\n')
            for e in bib_entry:
                if e is Tags.TY:
                    o.write(f"  Tags.{e.name}: {bib_entry[e]},\n")
                else:
                    o.write(f"  Tags.{e.name}: '{bib_entry[e]}',\n")
            o.write('}\n')
